import pandas as pd
import numpy as np

def _group_parser(indices, pop):
    # indices is a dictionary
    sel_ind = []
    for i in indices:
        if len(indices[i]) <= pop :
            sel_ind += indices[i]
        else:
            sel = list(np.random.choice(indices[i],size=pop,replace=False))
            sel_ind += sel
    return sel_ind

class Trimmer(object):
    """ cut unnecessary parts of the data
    
    Parameters
    ----------
    type: string, optional (default="margins")
        
        list of types:
        - random: remove part of data randomly
        - margins: cut from both top and bottom of data set
        - top: cut only top margin
        - bottom: cut only bottom margin
    
    sort: boolean, optional (default=True)
        If True data will be sorted by target values before trimming.
    
    cut: float in (0,1) range, optional (default=0.05)
        fraction of data set size to be trimmed.
    
    shuffle: Boolean, optional (default=True)
        To shuffle the data before sampling. Effective only if sort is False. 
    
    Attributes
    ----------
    Ncut: int
        number of removed data points
    
    selected_indices_: list
        Axis labels of data points that have been drawn randomly. Available only
        if type is 'random'.
          
    Returns
    -------
    data and target
    """
    def __init__(self, type="margins", sort = True,
                 cut = 0.05, shuffle = True):
        self.type = type
        self.sort = sort
        self.cut = cut
        self.shuffle = shuffle
        
    def fit_transform(self, data, target):
        """
        Fit the trimmer on df.
        """
        df = pd.concat([data, target], axis=1)
        if target.columns[0] in data.columns:
            cols = list(df.columns)
            col = 'target'
            while col in cols:
                col += 't'
            cols[-1] = col
            df.columns = cols
        else:
            col = target.columns[0]
        
        if self.sort == True:
            df.sort_values(col,axis=0,inplace=True)
            df.index = pd.Index(range(len(df)))
            data = df.iloc[:,:-1]
            target = pd.DataFrame(df.iloc[:,-1])
            
        elif self.shuffle == True:
            df = df.reindex(np.random.permutation(df.index))
            df.index = pd.Index(range(len(df)))            
            data = df.iloc[:,:-1]
            target = pd.DataFrame(df.iloc[:,-1])
                
        Nsamples = len(data)
        self.Ncut_ = int(self.cut * Nsamples) 
        if self.type == 'random':
            self.selected_indices_ = np.random.choice(range(0, Nsamples), Nsamples-self.Ncut_,replace=False)
            data = data.iloc[self.selected_indices_,:]
            data.index = pd.Index(range(len(data)))
            target = target.iloc[self.selected_indices_,:]
            target.index = pd.Index(range(len(target)))
            return data, target
        elif self.type == 'margins':
            Nhalfcut = self.Ncut_/2
            data = data[Nhalfcut:Nsamples-Nhalfcut]
            data.index = pd.Index(range(len(data)))
            target = target[Nhalfcut:Nsamples-Nhalfcut]
            target.index = pd.Index(range(len(target)))
            return data, target
        elif self.type == 'top':
            data = data[self.Ncut_:]
            data.index = pd.Index(range(len(data)))
            target = target[self.Ncut_:]
            target.index = pd.Index(range(len(target)))
            return data, target
        elif self.type == 'bottom':
            data = data[:Nsamples-self.Ncut_]
            data.index = pd.Index(range(len(data)))
            target = target[:Nsamples-self.Ncut_]
            target.index = pd.Index(range(len(target)))
            return data, target
        else:
            raise ValueError("Not a valid type")

class Uniformer(object):
    """ select a uniform size of groups of target values
    
    Parameters
    ----------
    bins: int or sequence of scalars or float, to be passed to pandas.cut
        If bins is an int, it defines the number of equal-width bins in the range of x. 
        However, in this case, the range of x is extended by 0.1% on each side to include 
        the min or max values of x. If bins is a sequence it defines the bin edges allowing
        for non-uniform bin width. No extension of the range of x is done in this case.
        If bins is a float, it defines the width of bins.
    
    right: bool, optional, default True
        Indicates whether the bins include the rightmost edge or not. 
        If right == True (by default), then the bins [1,2,3,4] indicate (1,2], (2,3], (3,4].
    
    include_lowest: bool, optional, default False
        Whether the first interval should be left-inclusive or not.
    
    bin_pop: int or float, optional, default 0.5
        bin_pop defines the maximum population of selected samples from each group(bin).
        If bin_pop is an int, it defines the maximum number of samples to be drawn from each group.
        A float value for bin_pop defines the fraction of the maximum population of groups as the 
        maximum size of selections from each group.
        
    substitute: str('mean','lower,'upper') or sequence of scalars, optional, default None
        If substitute is one of the choices of 'mean', 'lower' or 'upper' strings, 
        target values will be substitute with mean bin edges, lower bin edge or 
        upper bin edge, respectively. If bins is a sequence, it defines the target 
        value for bins. If None, no substitute will happen and original target 
        values would be passed out.
    
    Attributes
    ----------
    groups_: dataframe, shape (n_targets, n_bins, n_ranges)
        return groups info: target values, bin labels, range of bins
    
    grouped_indices_: dictionary
        A dict whose keys are the group labels and corresponding values being the 
        axis labels belonging to each group.
    
    selected_indices_: list
        axis labels of data points that are drawn.
    
    Returns
    -------
    data and target
    """
    def __init__(self, bins, bin_pop = 0.5, right = True, include_lowest = True,
                 substitute = None):
        self.bins = bins
        self.bin_pop = bin_pop
        self.right = right
        self.include_lowest = include_lowest
        self.substitute = substitute
        
    def fit_transform(self, data, target):
        """
        Fit the uniformer on df.
        """
        # pandas.cut
        col = target.columns[0]
        if type(self.bins) == int:
            bined = pd.cut(target[col], self.bins, right = self.right, retbins=True, labels=False,include_lowest=self.include_lowest)
        elif type(self.bins) == float:
            bins = int(max(target[col]) - min(target[col]) / self.bins)
            bined = pd.cut(target[col], bins, right = self.right, retbins=True, labels=False,include_lowest=self.include_lowest)
        else:
            bined = pd.cut(target[col], self.bins, right = self.right, retbins=True, labels=False,include_lowest=self.include_lowest)
        
        if self.right:
            ranges = ['(%f,%f]'%(bined[1][i],bined[1][i+1]) for i in bined[0]]
            if self.include_lowest:
                ranges[0] = '[' + ranges[0][1:]
        else:
            ranges = ['(%f,%f)'%(bined[1][i],bined[1][i+1]) for i in bined[0]]
            if self.include_lowest:
                ranges[0] = '[' + ranges[0][1:]
        
        # pandas.groupby
        self.groups_ = pd.DataFrame()
        self.groups_['target'] = target[col]
        self.groups_['bins'] = bined[0]
        self.groups_['ranges'] = ranges
        self.groups_.sort_values('bins',axis=0,inplace=True)
        self.grouped_indices_ = self.groups_.groupby('bins').groups
         
        if type(self.bin_pop) == int:
            pop = self.bin_pop    
        elif type(self.bin_pop) == float:
            pop = int(self.bin_pop * max([len(self.grouped_indices_[i]) for i in self.grouped_indices_]))
        else:
            raise ValueError("Wrong format of bin_pop: must be int or float")              
        
        self.selected_indices_ = _group_parser(self.grouped_indices_, pop)
        
        data = data.iloc[self.selected_indices_,:]
        data.index = pd.Index(range(len(data)))        
        if self.substitute == None :
            target = target.iloc[self.selected_indices_,:]        
            target.index = pd.Index(range(len(target)))
        else:
            if self.substitute == 'lower':
                new_target_values = [bined[1][bined[0][i]] for i in self.selected_indices_]
            elif self.substitute == 'upper':
                new_target_values = [bined[1][bined[0][i]+1] for i in self.selected_indices_]    
            elif self.substitute == 'mean':
                new_target_values = [np.mean([bined[1][bined[0][i]], bined[1][bined[0][i]+1]])for i in self.selected_indices_]
            else:
                new_target_values = list(self.substitute)
        
            target = pd.DataFrame(new_target_values, columns=[col])        
        
        return data, target




# coding=utf-8
from numpy.linalg import norm
import numpy as np
from .voronoi.VoronoiTessellationCalculator import \
    VoronoiTessellationCalculator

class VoronoiCellBasedAnalysis:
    """Class to perform structure analysis based on the Voronoi tessellation
    method.

    Attributes
    ----------
    structure : Cell
        Link to structure being evaluated.
    cells : array-like
        Voronoi cells for each atom in structure.
    radical : bool
        Whether to use a radical Voronoi tessellation.

    """

    def __init__(self, radical=None, old_tessellation=None,
                 new_structure=None):
        """Function to initialize a Voronoi cell analyzer.

        Create a new instance of the analysis toolkit based on a structure
        that has an identical tessellation. This occurs when the two
        structures have identical lattice parameters and atomic positions,
        but different identities of elements on those sites.

        Parameters
        new_structure : Cell
            New structure.
        radical : bool
            Whether to use a radical Voronoi tessellation.
        old_tessellation : VoronoiCellBasedAnalysis
            Tessellation of original structure.

        Raises
        ------
        Exception
            If structures have different number of atoms.
            If basis vectors of structures are different.
            If any atom position is different
        """

        # Link to structure being evaluated.
        self.structure = None

        # Voronoi cells for each atom in structure.
        self.cells = None

        # Whether to use a radical Voronoi tessellation.
        self.radical = False

        if radical is not None:
            self.radical = radical
        else:
            # Check whether structures are similar.
            old_structure = old_tessellation.structure
            if old_structure.n_atoms() != new_structure.n_atoms():
                raise Exception("Structures have different number of atoms.")
            if norm(old_structure.get_basis() - new_structure.get_basis()) > \
                    1e-6:
                raise Exception("Basis vectors of structures are different.")
            for a in range(old_structure.n_atoms()):
                old_atom_pos = old_structure.get_atom(a).get_position()
                new_atom_pos = new_structure.get_atom(a).get_position()
                if not np.allclose(old_atom_pos, new_atom_pos, rtol=1e-6):
                    raise Exception("Atom # {} position is different".format(
                        a))

            # Create new instance.
            self.cells = old_tessellation.cells
            self.structure = new_structure
            self.radical = old_tessellation.radical

    def precompute(self):
        """Function to perform any kind of computations that should be
        performed only once.

        Raises
        ------
        Exception
            If geometry is invalid.
        """
        self.cells = VoronoiTessellationCalculator.compute(self.structure,
                                                           self.radical)
        for cell in self.cells:
            if not cell.geometry_is_valid():
                self.cells = None
                raise Exception("Invalid geometry.")

    def analyze_structure(self, s):
        """Function to analyze a specific structure. Once this completes,
        it is possible to retrieve results out of this object.

        Parameters
        ----------
        s : Cell
            Structure to be analyzed.


        """
        self.structure = s
        self.precompute()

    def recompute(self):
        """Function to recompute structural information.

        """
        self.precompute()

    def tessellation_is_converged(self):
        """Function to check whether the tessellation of this structure was
        successful.

        """
        return self.cells is not None

    def get_effective_coordination_numbers(self):
        """Function to get the effective coordination number.

        Defined as N_eff = 1 / sum[(f_i / SA_i)^2] where f_i is the area of
        face i.

        Parameters
        ----------

        Returns
        -------
        output : list
            Effective coordination number for each atom.
        """

        # Compute the coordination number.
        output = [1.0 / sum([(f_i.get_area() / cell.get_surface_area()) ** 2
                    for f_i in cell.get_faces()]) for cell in self.cells]
        return output

    def face_count_average(self):
        """Function to get the average number of faces on all cells.

        Returns
        -------
        output : float
            Average number.

        """
        return np.average([cell.n_faces() for cell in self.cells])

    def face_count_variance(self):
        """Function to get the variance in face count.

        Returns
        -------
        output : float
            Mean absolute deviation in coordination number.
        """
        avg = self.face_count_average()
        return np.mean([abs(cell.n_faces() - avg) for cell in self.cells])

    def face_count_minimum(self):
        """Function to get the minimum face count of all cells.

        Returns
        -------
        output : float
            Minimum.
        """
        return min([cell.n_faces() for cell in self.cells])

    def face_count_maximum(self):
        """Function to get the maximum face count of all cells.

        Parameters
        ----------

        Returns
        -------
        output : float
            Maximum.
        """
        return max([cell.n_faces() for cell in self.cells])

    def get_unique_polyhedron_shapes(self):
        """Function to get a list of all unique polyhedron shapes.

        Returns
        -------
        output : list
            Set of polyhedron shapes.
        """
        output = []
        for cell in self.cells:
            x = cell.get_polyhedron_shape()
            if x not in output:
                output.append(x)
        return output

    def volume_variance(self):
        """Function to compute the mean absolute deviation in the volume of each
        cell.

        Returns
        -------
        output : float
            Variance of cell volume fraction of all cells.
        """
        avg_volume = self.structure.volume() / self.structure.n_atoms()
        return np.mean([abs(self.cells[i].get_volume() - avg_volume) for i
                           in range(self.structure.n_atoms())])

    def volume_fraction_minimum(self):
        """Function to compute the fraction of cell volume occupied by the
        smallest Voronoi cell.

        Returns
        -------
        output : float
            Volume fraction of the smallest cell.
        """
        return min([cell.get_volume() for cell in self.cells]) / \
               self.structure.volume()

    def volume_fraction_maximum(self):
        """Function to compute the fraction of cell volume occupied by the
        largest Voronoi cell.

        Returns
        -------
        output : float
            Volume fraction of the largest cell.

        """
        return max([cell.get_volume() for cell in self.cells]) / \
               self.structure.volume()

    def max_packing_efficiency(self):
        """Function to compute the maximum packing efficiency assuming atoms are
        hard spheres.

        Algorithm:
        1. For each cell in the Voronoi tessellation of this cell, determine the
        minimum distance from the cell center to a face. This marks the maximum
        atomic radius.
        2. Compute the total volume represented by those maximally-sized atoms.
        3. Packing efficiency is atom volume divided by cell volume.

        Returns
        -------
        output : float
            Maximum packing efficiency.

        """
        atom_vol = sum([min_dist ** 3 for min_dist in [min([
            face.get_face_distance() for face in cell.get_faces()]) for cell
            in self.cells]])
        atom_vol *= 4.0 * np.math.pi / 3.0
        return atom_vol / self.structure.volume()

    def get_neighbor_ordering_parameters(self, shell, weighted):
        """Function to compute the Warren-Cowley short range ordering parameters
        for each atom.

        alpha_{s, i} = 1 - n_{A, s} / (x_A * n_s)
        where n_{A, s} is the number of atoms of type A in shell s, x_A is
        the composition of atom A and n_s is the number of atoms in shell s.

        Optionally, one can weight the contributions of each neighbor based on
        the path weights. See VoronoiCell.get_neighbors_by_walks for further
        discussion.

        Parameters
        ----------
        shell : int
            Index of nearest neighbor shell.
        weighted : bool
            Whether to compute the weighted ordering parameters.

        Returns
        -------
        output : array-like
            Ordering parameters for each atom in cell.

        References
        ----------
        .. [1] J. M. Cowley, "An Approximate Theory of Order in Alloys,
        " Physical Review, vol. 77, no. 5, pp. 669--675, Mar. 1950.
        """
        n_atoms = self.structure.n_atoms()
        n_types = self.structure.n_types()
        output = np.zeros((n_atoms, n_types), dtype=float)

        # Compute the composition.
        x = np.zeros(n_types, dtype=float)
        for atom in self.structure.get_atoms():
            x[atom.get_type()] += 1

        x /= n_atoms

        # Compute the WC parameters for each atom.
        for a in range(n_atoms):
            if weighted:
                # Get the neighbors and weights.
                neighbor_shell = self.cells[a].get_neighbors_by_walks(
                    self.cells, shell)

                # Compute the actual weight of each type.
                n_a = np.zeros(n_types, dtype=float)
                for k,v in neighbor_shell.iteritems():
                    n_a[k.get_atom().get_type()] += v

                # Compute the WC parameters.
                for t in range(n_types):
                    if x[t] != 0:
                        output[a][t] = 1 - n_a[t] / x[t]
            else:
                # Get the neighboring atoms.
                neighbors = self.cells[a].get_neighbor_shell(self.cells, shell)

                # Compute the actual number of each type.
                n_a = np.zeros(n_types, dtype=float)
                for neighbor in neighbors:
                    n_a[neighbor.get_atom().get_type()] += 1

                # Compute the WC parameters.
                l_n = len(neighbors)
                for t in range(n_types):
                    if x[t] != 0:
                        output[a][t] = 1 - n_a[t] / l_n / x[t]
        return output

    def warren_cowley_ordering_magnitude(self, shell, weighted):
        """Function to compute the mean deviation in the Warren-Cowley
        parameter for each type for site from 0.

        Consider this as a measure of how "ordered" a structure is. Computed
        as the average of the average the absolute values of the WC
        parameters for each type for each site:
        Sum_{i,j} 1 / [Number of atoms in structure] / [Number of types] * |
        [ WC parameter for type i about atom j] |

        Parameters
        ----------
        shell : int
            Index of neighbor shell (e.g. 1st shell = 1).
        weighted : bool
            Whether to weigh ordering parameters by face area.

        Returns
        -------
        output : float
            Computed value.

        """

        # Get the WC ordering parameters.
        wc = self.get_neighbor_ordering_parameters(shell, weighted)
        return np.sum(np.abs(wc)) / self.structure.n_atoms() / \
               self.structure.n_types()

    def compute_shape_dissimilarity(self, this_shape, reference_shape):
        """Function to compute the similarity of the shape of a cell to a
        reference.

        Computed as the difference between the number of faces with a certain
        number of edges for between a shape and the reference for each type
        of face (defined by the number of edges) divided by the total number
        of faces in the reference shape.

        Example: Reference shape has 12 square faces, this shape has 11
        square faces and two triangular faces. There are three different faces
        (i.e. one missing square face and two extraneous triangular faces).
        Therefore the dissimilarity is: 3 / 12 = 25%.

        Parameters
        ----------
        this_shape : dict
            Shape to be compared.
        reference_shape : dict
            Shape to compare against.

        Returns
        -------
        output : float
            Dissimilarity figure (0 means identical, can be > 1).

        """
        n_diff = 0
        n_faces = 0

        # For all shapes that in reference pattern.
        for type, count in reference_shape.iteritems():
            n_faces += count
            if type in this_shape:
                n_diff += abs(count - this_shape[type])
            else:
                n_diff += count

        # For all types in this shape, but not reference pattern.
        types_to_eval = set(this_shape.keys())
        types_to_eval -= set(reference_shape.keys())
        for type in types_to_eval:
            n_diff += this_shape[type]

        return float(n_diff) / n_faces

    def mean_sc_dissimilarity(self):
        """Function to get how dissimilar, on average, coordination polyhedra
        are from super cell.

        Returns
        -------
        output : float
            Average shape dissimilarity from sc.
        """

        # Make reference.
        ref = {4 : 6}

        # Return mean dissimilarity.
        return np.mean([self.compute_shape_dissimilarity(self.cells[
            i].get_polyhedron_shape(), ref) for i in range(
            self.structure.n_atoms())])

    def mean_bcc_dissimilarity(self):
        """Function to get how dissimilar, on average, coordination polyhedra
        are from bcc.

        Returns
        -------
        output : float
            Average shape dissimilarity from bcc.
        """

        # Make reference.
        ref = {6: 8, 4: 6}

        # Return mean dissimilarity.
        return np.mean([self.compute_shape_dissimilarity(self.cells[
            i].get_polyhedron_shape(), ref) for i in range(
            self.structure.n_atoms())])

    def mean_fcc_dissimilarity(self):
        """Function to get how dissimilar, on average, coordination polyhedra
        are from fcc.

        Returns
        -------
        output : float
            Average shape dissimilarity from fcc.

        """

        # Make reference.
        ref = {4: 12}

        # Return mean dissimilarity.
        return np.mean([self.compute_shape_dissimilarity(self.cells[
            i].get_polyhedron_shape(), ref) for i in range(
            self.structure.n_atoms())])

    def neighbor_property_differences(self, property, shell):
        """Function to compute the face-size-weighted difference between
        properties of an atom and its neighbors.

        Computed as:
        sum_i [ Size of face between atom and neighbor i ] * |
        [property of atom] - [property neighbor] | / [ surface area ]

        For neighbors in the 2nd or greater shell,
        VoronoiCell.get_extended_faces is used to find the unique faces
        between the atoms in the N - 1 shell and the Nth shell. So, each Nth
        neighbor atom might have multiple faces and its total "size of face"
        will be defined as the sum of the area of these faces.

        Parameters
        ----------
        property : array-like
            List of property for each atom type.
        shell : int
            Shell to be considered (1 == 1st nearest neighbor shell).

        Returns
        -------
        output : array-like
            Property difference value for each neighbor.

        """

        # Get lookup table of face areas and types.
        neighbors = self.get_neighbor_shell_weights(shell)
        face_types = neighbors[0]
        face_areas = neighbors[1]

        n_atoms = self.structure.n_atoms()
        # Compute average difference for each cell.
        output = np.zeros(n_atoms, dtype=float)
        for a in range(n_atoms):
            # Get the property of this atom.
            my_value = property[self.structure.get_atom(a).get_type()]
            output[a] = np.average([abs(my_value - property[face_types[a][f]])
                        for f in range(len(face_areas[a]))],
                                   weights=face_areas[a])

        return output

    def get_neighbor_shell_weights(self, shell):
        """Function to get the types and weights on neighbors in each shell.

        Uses the path weight argument described in
        VoronoiCell.get_neighbors_by_walks.

        Parameters
        ----------
        shell : int
            Shell being considered. 1 corresponds to the 1st NN,
            2 corresponds to the polyhedron formed by an atom and its 1st shell
            neighbors, etc.

        Returns
        -------
        output : array-like
            Pair of neighbor types and weights.

        """
        n_atoms = self.structure.n_atoms()
        types = np.zeros((n_atoms, ), dtype=np.ndarray)
        weights = np.zeros((n_atoms, ), dtype=np.ndarray)

        for a in range(n_atoms):
            # Get extended faces.
            neighbors = self.cells[a].get_neighbors_by_walks(self.cells, shell)
            l_n = len(neighbors)
            # Store their areas.
            types[a] = np.array([k.get_atom().get_type() for k in
                                 neighbors.keys()], dtype=int)
            weights[a] = np.array(neighbors.values(), dtype=float)

        return [types, weights]

    def neighbor_property_variances(self, property, shell):
        """Function to get the weighted variance between the properties of each
        atom's neighbors.

        Parameters
        ----------
        property : array-like
            List of property for each type.
        shell : int
            Shell being considered. 1 corresponds to the 1st NN,
            2 corresponds to the polyhedron formed by an atom and its 1st shell
            neighbors, etc.

        Returns
        -------
        output : array-like
            Variance in property of each atom for each neighbor.

        Raises
        ------
        Exception
            If wrong number of property values provided.
        """

        if len(property) != self.structure.n_types():
            raise Exception("Wrong number of property values provided.")

        neighbors = self.get_neighbor_shell_weights(shell)
        types = neighbors[0]
        weights = neighbors[1]

        n_atoms = self.structure.n_atoms()
        # Compute average difference for each cell.
        output = np.zeros(n_atoms, dtype=float)

        for a in range(n_atoms):
            l_t = len(types[a])

            # Get the properties of all neighbors.
            face_props = np.array([property[types[a][f]] for f in range(
                l_t)], dtype=float)

            # Compute the face-weighted mean.
            mean = np.average(face_props, weights=weights[a])

            # Compute the face-weighted variance.
            x = face_props - mean * np.ones(l_t)
            output[a] = np.average(x ** 2, weights=weights[a])
            # print a, mean, output[a]

        return output

    def bond_lengths(self):
        """Function to get the bond lengths for each atom.

        Returns
        -------
        output : array-like
            Bond lengths for each atom in Cartesian units.
        """
        n_atoms = self.structure.n_atoms()
        output = np.array([self.cells[a].get_neighbor_distances() for a in
                           range(n_atoms)])
        return output

    def mean_bond_lengths(self):
        """Function to compute mean bond length for each cell.

        Bond length for each neighbor is weighted by face size.

        Returns
        -------
        output : float
            Mean bond lengths.

        """
        n_atoms = self.structure.n_atoms()
        output = np.array([sum([face.get_area() * face.get_neighbor_distance()
                for face in self.cells[a].get_faces()]) /
                self.cells[a].get_surface_area() for a in range(n_atoms)])
        return output

    def bond_length_variance(self, mean_lengths):
        """Function to compute variance in bond length for each face.

        Computed as the mean absolute deviation of each neighbor distance
        from the mean neighbor distance, weighted by face area.

        Parameters
        ----------
        mean_lengths : float
            Mean bond length for each cell.

        Returns
        -------
        output : array-like
            Bond length variance for each cell.

        """
        n_atoms = self.structure.n_atoms()
        output = np.array([sum([face.get_area() *
                abs(face.get_neighbor_distance() - mean_lengths[a])
                    for face in self.cells[a].get_faces()]) /
                    self.cells[a].get_surface_area() for a in range(n_atoms)])
        return output


"""
This module is adapted entirely from Magpie (https://bitbucket.org/wolverton/magpie).
If you are using this module, please cite Magpie as:
 L. Ward, A. Agrawal, A. Choudhary, and C. Wolverton, "A general-purpose machine learning framework for predicting properties of inorganic materials," npj Computational Materials, vol. 2, no. 1, Aug. 2016.
For more information regarding the python version of Magpie, please see https://github.com/ramv2/magpie_python.
The cheml.chem.magpie_python module includes (please click on links adjacent to function names for more information):
    - APEAttributeGenerator: :func:`~cheml.chem.magpie_python.APEAttributeGenerator`
    - ChargeDependentAttributeGenerator: :func:`~cheml.chem.magpie_python.ChargeDependentAttributeGenerator`
    - ElementalPropertyAttributeGenerator: :func:`~cheml.chem.magpie_python.ElementalPropertyAttributeGenerator`
    - ElementFractionAttributeGenerator: :func:`~cheml.chem.magpie_python.ElementFractionAttributeGenerator`
    - ElementPairPropertyAttributeGenerator: :func:`~cheml.chem.magpie_python.ElementPairPropertyAttributeGenerator`
    
    - GCLPAttributeGenerator: :func:`~cheml.chem.magpie_python.GCLPAttributeGenerator`
    - IonicCompoundProximityAttributeGenerator: :func:`~cheml.chem.magpie_python.IonicCompoundProximityAttributeGenerator`
    - IonicityAttributeGenerator: :func:`~cheml.chem.magpie_python.IonicityAttributeGenerator`
    - MeredigAttributeGenerator: :func:`~cheml.chem.magpie_python.MeredigAttributeGenerator`
    - StoichiometricAttributeGenerator: :func:`~cheml.chem.magpie_python.StoichiometricAttributeGenerator`
    
    - ValenceShellAttributeGenerator: :func:`~cheml.chem.magpie_python.ValenceShellAttributeGenerator`
    - YangOmegaAttributeGenerator: :func:`~cheml.chem.magpie_python.YangOmegaAttributeGenerator`
    - APRDFAttributeGenerator: :func:`~cheml.chem.magpie_python.APRDFAttributeGenerator`
    - ChemicalOrderingAttributeGenerator: :func:`~cheml.chem.magpie_python.ChemicalOrderingAttributeGenerator`
    - CoordinationNumberAttributeGenerator: :func:`~cheml.chem.magpie_python.CoordinationNumberAttributeGenerator`
    
    - CoulombMatrixAttributeGenerator: :func:`~cheml.chem.magpie_python.CoulombMatrixAttributeGenerator`
    - EffectiveCoordinationNumberAttributeGenerator: :func:`~cheml.chem.magpie_python.EffectiveCoordinationNumberAttributeGenerator`
    - LatticeSimilarityAttributeGenerator: :func:`~cheml.chem.magpie_python.LatticeSimilarityAttributeGenerator`
    - LocalPropertyDifferenceAttributeGenerator: :func:`~cheml.chem.magpie_python.LocalPropertyDifferenceAttributeGenerator`
    - LocalPropertyVarianceAttributeGenerator: :func:`~cheml.chem.magpie_python.LocalPropertyVarianceAttributeGenerator`

    - PackingEfficiencyAttributeGenerator: :func:`~cheml.chem.magpie_python.PackingEfficiencyAttributeGenerator`
    - PRDFAttributeGenerator: :func:`~cheml.chem.magpie_python.PRDFAttributeGenerator`
    - StructuralHeterogeneityAttributeGenerator: :func:`~cheml.chem.magpie_python.StructuralHeterogeneityAttributeGenerator`
    - CompositionEntry: :func:`~cheml.chem.magpie_python.CompositionEntry`
    - CrystalStructureEntry: :func:`~cheml.chem.magpie_python.CrystalStructureEntry`
"""
__all__ = [
    'APEAttributeGenerator',
    'ChargeDependentAttributeGenerator',
    'ElementalPropertyAttributeGenerator',
    'ElementFractionAttributeGenerator',
    'ElementPairPropertyAttributeGenerator',
    
    'GCLPAttributeGenerator',
    'IonicCompoundProximityAttributeGenerator',
    'IonicityAttributeGenerator',
    'MeredigAttributeGenerator',
    'StoichiometricAttributeGenerator',
    
    'ValenceShellAttributeGenerator',
    'YangOmegaAttributeGenerator',
    'APRDFAttributeGenerator',
    'ChemicalOrderingAttributeGenerator',
    'CoordinationNumberAttributeGenerator',
    
    'CoulombMatrixAttributeGenerator',
    'EffectiveCoordinationNumberAttributeGenerator',
    'LatticeSimilarityAttributeGenerator',
    'LocalPropertyDifferenceAttributeGenerator',
    'LocalPropertyVarianceAttributeGenerator',

    'PackingEfficiencyAttributeGenerator',
    'PRDFAttributeGenerator',
    'StructuralHeterogeneityAttributeGenerator',
    'CompositionEntry',
    'CrystalStructureEntry',
]
from attributes.generators.composition.APEAttributeGenerator import \
    APEAttributeGenerator
from attributes.generators.composition.ChargeDependentAttributeGenerator \
    import ChargeDependentAttributeGenerator
from attributes.generators.composition.ElementalPropertyAttributeGenerator \
    import ElementalPropertyAttributeGenerator
from attributes.generators.composition.ElementFractionAttributeGenerator \
    import ElementFractionAttributeGenerator
from attributes.generators.composition.ElementPairPropertyAttributeGenerator \
    import ElementPairPropertyAttributeGenerator


from attributes.generators.composition.GCLPAttributeGenerator import \
    GCLPAttributeGenerator
from attributes.generators.composition\
    .IonicCompoundProximityAttributeGenerator import \
    IonicCompoundProximityAttributeGenerator
from attributes.generators.composition.IonicityAttributeGenerator import \
    IonicityAttributeGenerator
from attributes.generators.composition.MeredigAttributeGenerator import \
    MeredigAttributeGenerator
from attributes.generators.composition.StoichiometricAttributeGenerator \
    import StoichiometricAttributeGenerator


from attributes.generators.composition.ValenceShellAttributeGenerator import \
    ValenceShellAttributeGenerator
from attributes.generators.composition.YangOmegaAttributeGenerator import \
    YangOmegaAttributeGenerator
from attributes.generators.crystal.APRDFAttributeGenerator import \
    APRDFAttributeGenerator
from attributes.generators.crystal.ChemicalOrderingAttributeGenerator import \
    ChemicalOrderingAttributeGenerator
from attributes.generators.crystal.CoordinationNumberAttributeGenerator \
    import CoordinationNumberAttributeGenerator


from attributes.generators.crystal.CoulombMatrixAttributeGenerator import \
    CoulombMatrixAttributeGenerator
from attributes.generators.crystal\
    .EffectiveCoordinationNumberAttributeGenerator import \
    EffectiveCoordinationNumberAttributeGenerator
from attributes.generators.crystal.LatticeSimilarityAttributeGenerator import\
    LatticeSimilarityAttributeGenerator
from attributes.generators.crystal.LocalPropertyDifferenceAttributeGenerator \
    import  LocalPropertyDifferenceAttributeGenerator
from attributes.generators.crystal.LocalPropertyVarianceAttributeGenerator \
    import LocalPropertyVarianceAttributeGenerator


from attributes.generators.crystal.PackingEfficiencyAttributeGenerator import\
    PackingEfficiencyAttributeGenerator
from attributes.generators.crystal.PRDFAttributeGenerator import \
    PRDFAttributeGenerator
from attributes.generators.crystal.StructuralHeterogeneityAttributeGenerator \
    import StructuralHeterogeneityAttributeGenerator

from data.materials.CompositionEntry import CompositionEntry
from data.materials.CrystalStructureEntry import CrystalStructureEntry

# __all__ = ["attributes", "data", "models", "test", "utility", "vassal"]



import itertools
import numpy as np
from ...data.materials.CompositionEntry import CompositionEntry

class OxidationStateGuesser:
    """Class to predict the likely oxidation states of a material, given its
    input composition.

    Attributes
    ----------
    electronegativity : array-like
        A list of electronegativity values (float).
    oxidation_states : array-like
        A 2-D numpy array containing the property values for all the elements.

    """
    electronegativity = np.zeros(0)
    oxidationstates = np.zeros(0, dtype=object)

    def set_electronegativity(self, values):
        """Function to set the electronegativity values.

        Parameters
        ----------
        values : array-like
            Numpy array containing electronegativity values for all the
            elements.

        Returns
        -------

        """
        self.electronegativity = values

    def set_oxidationstates(self, values):
        """Function to set the oxidation states values.

        Parameters
        ----------
        values : array-like
            2-D numpy array containing oxidation states values for all the
            elements.

        Returns
        -------

        """
        self.oxidationstates = values

    def get_possible_states(self, entry):
        """Function to compute all the possible oxidation states of a material,
        given its input composition.

        The function works by finding all
        combinations of non-zero oxidation states for each element, computing
        which are the most reasonable, and finding which of those have minimum
        value of
        sum_{i,j} (chi_i - chi_j)*(c_i - c_j) for i < j
        where chi_i is the electronegativity and c_i is the oxidation. This
        biases the selection towards the more electronegative elements being
        more negatively charged.

        Parameters
        ----------
        entry : CompositionEntry
            A CompositionEntry object.

        Returns
        -------
        output : array-like
            A numpy array containing the list of possible oxidation states
            arranged in the order mentioned above.

        Raises
        ------
        ValueError
            If input is empty.
            If input is not a CompositionEntry object.
            If electronegativity or oxidationstates haven't been set.

        """

        # Make sure entry is not empty.
        if not entry:
            raise ValueError("Input argument cannot be empty. Please pass a "
                             "valid argument.")


        # Make sure entry is of type CompositionEntry.
        if not isinstance(entry, CompositionEntry):
            raise ValueError("Entry must be of type CompositionEntry.")

        # Make sure electronegativity and oxidation states are not empty.
        if not self.electronegativity.size or not self.oxidationstates.size:
            raise ValueError("Electronegativity or OxidationStates values are "
                             "not initialized. Set them and try again.")

        # Initialize list of possible states.
        possible_states = []

        # Get element ids and fractions.
        elem_ids = entry.get_element_ids()
        elem_fracs = entry.get_element_fractions()
        if len(elem_ids) == 1:
            return np.asarray([])

        # List of all states.
        states = []
        for id in elem_ids:
            states.append(self.oxidationstates[id])

        # Generate all combinations of those charge states, only store the
        # ones that are charge balanced.
        for state in itertools.product(*states):
            charge = np.dot(state, elem_fracs)
            # If charge is balanced, add state to the list of possible states.
            if abs(charge) < 1E-6:
                possible_states.append(list(state))

        if len(possible_states) < 2:
            return np.asarray(possible_states)

        # Compute the summation mentioned in the function description.
        rankVal = np.zeros(len(possible_states))
        for s in range(len(possible_states)):
            state = possible_states[s]
            tmp_val = 0.0
            for i in range(len(state)):
                for j in range(i+1,len(state)):
                    tmp_val += (self.electronegativity[elem_ids[i]] -
                                self.electronegativity[elem_ids[j]]) * (
                        state[i] - state[j])
            rankVal[s] = tmp_val

        # Order them based on electronegativity rank.
        output = [ps for i, ps in sorted(zip(rankVal, possible_states))]
        return np.asarray(output)


import pandas as pd
import numpy as np
import types
from ....data.materials.CrystalStructureEntry import CrystalStructureEntry
from ....data.materials.util.LookUpData import LookUpData

class LocalPropertyDifferenceAttributeGenerator:
    """Class to compute attributes based on the difference in elemental
    properties between neighboring atoms.

    Attributes
    ----------
    elemental_properties : list
        Elemental properties to be associated with this class for the
        generation of features.
    shells : array-like
        Shells to consider. A list of int values.
    attr_name : str
        Property Name.

    Notes
    -----
    For an atom, its "local property difference" is computed by:

    .. math:: \displaystyle\frac{\sum_n f_n * \left|p_{atom} - p_n\right|}{
    \sum_n f_n}

    where :math: `f_n` is the area of the face associated with neighbor
    :math: `n, p_{atom}` is the elemental property of the central atom,
    and :math: `p_n` is the elemental property of the neighbor atom.

    For shells past the 1st nearest neighbor shell, the neighbors are
    identified by finding all of the unique faces on the outside of the
    polyhedron formed by the previous neighbor shell. This list of faces
    will faces corresponding to all of the atoms in the desired shell and the
    total weight for each atom is defined by the total area of the faces
    corresponding to that atom (there may be more than one).

    By default, this class considers the only the 1st nearest neighbor shell.

    This parameter is computed for all elemental properties stored in
    Composition Entry ElementalProperties.

    """

    def __init__(self, shells=None):
        """Function to create instance and initialize fields.

        Parameters
        ----------
        shells : array-like
            Shells to be considered. A list of int values.

        """

        # Elemental properties used to generate attributes.
        self.elemental_properties = []

        # Shells to consider.
        self.shells = [1] if shells is None else list(set(shells))

        # Property Name.
        self.attr_name = "NeighDiff"

        # Property description (used in description output).
        # self.attr_description = "difference between the elemental
        # properties between an atom and neighbors"

    def clear_shells(self):
        """Function to clear the list of shells.

        """

        self.shells = []

    def add_shell(self, shell):
        """Function to add shell to the list used when computing attributes.

        Parameters
        ----------
        shells : int
            Index of nearest-neighbor shell.

        Raises
        ------
        ValueError
            If shell is negative.

        """

        if shell <= 0:
            raise ValueError("Shell index must be > 0.")
        if shell not in self.shells:
            self.shells.append(shell)

    def add_shells(self, shells):
        """Function to add a list of shells to be used when computing
        attributes.

        Parameters
        ----------
        shells : array-like
            Shells to be considered. A list of int values.

        """

        for s in shells:
            self.add_shell(s)

    def add_elemental_property(self, prop):
        """Function to add an elemental property to `self.elemental_properties`
        in order to be used to compute features.

        Parameters
        ----------
        property : str
            Property to be added.

        """

        if prop not in self.elemental_properties:
            self.elemental_properties.append(prop)

    def add_elemental_properties(self, properties):
        """Function to provide a list of elemental properties to be used to
        compute features.

        Parameters
        ----------
        properties : array-like
            Properties to be included. A list of strings containing property
            names.

        """

        for p in properties:
            self.add_elemental_property(p)

    def remove_elemental_property(self, property):
        """Function to remove an elemental property from
        `self.elemental_properties`.

        Parameters
        ----------
        property : str
            Property to be removed.

        """

        if property in self.elemental_properties:
            self.elemental_properties.remove(property)

    def remove_elemental_properties(self, properties):
        """Function to remove a list of elemental properties from the list of
        elemental properties.

        Parameters
        ----------
        properties : array-like
            Properties to be removed. A list of strings containing property
            names.

        """

        for prop in properties:
            self.remove_elemental_property(prop)

    def clear_elemental_properties(self):
        """Function to clear all the elemental properties.

        """

        self.elemental_properties = []

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Crystal structures for which features are to be generated. A list
            of CrystalStructureEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CrystalStructureEntry instances.

        """

        # Initialize list of feature values for pandas data frame.
        feat_values = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")
        elif (entries and not isinstance(entries[0], CrystalStructureEntry)):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")

        # Insert header names here.
        prefix = ["mean_", "var_", "min_", "max_", "range_"]
        feat_headers =[p+self.attr_name+"_shell{}".format(s)+"_"+prop for s
                       in self.shells for prop in self.elemental_properties
                       for p in prefix]

        # Generate features.
        l_fh = len(feat_headers)
        for entry in entries:
            tmp_list = []
            # Get the Voronoi tessellation.
            try:
                voro = entry.compute_voronoi_tessellation()
            except Exception:
                tmp_list = [np.nan] * l_fh  # If tessellation fails.
                feat_values.append(tmp_list)
                continue

            # Get the elements corresponding to each type.
            st = entry.get_structure()
            elem_index = np.array([LookUpData.element_names.index(
                            st.get_type_name(i)) for i in range(st.n_types())])

            # Loop through each shell.
            for shell in self.shells:
                # Loop through each elemental property.
                for prop in self.elemental_properties:
                    # Get properties for elements in this structure.
                    lookup_table = LookUpData.load_property(prop)
                    prop_values = np.array([lookup_table[elem_index[i]] for i
                                            in range(len(elem_index))])

                    # Compute neighbor differences.
                    neigh_diff = self.get_atom_properties(voro, shell,
                                                          prop_values)
                    mean = np.mean(neigh_diff)
                    tmp_list.append(mean)
                    tmp_list.append(np.mean([abs(x - mean) for x in
                                             neigh_diff]))
                    min_ = np.min(neigh_diff)
                    max_ = np.max(neigh_diff)
                    tmp_list.append(min_)
                    tmp_list.append(max_)
                    tmp_list.append(max_ - min_)

            feat_values.append(tmp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features

    def get_atom_properties(self, voro, shell, prop_values):
        """Function to compute the properties of a certain neighbor cell for
        each atom, given the Voronoi tessellation and properties of each atom
        type.

        Parameters
        ----------
        voro : VoronoiCellBasedAnalysis
            Analysis tool.
        shell : int
            Index of shell.
        prop_values : array-like
            Properties of each atom type. A list or NumPy array of float values.

        Returns
        -------
        output : array-like
            Properties of each atom. A list or NumPy array of float values.

        """

        output = voro.neighbor_property_differences(prop_values, shell)
        return output


import os
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt



class decorator(object):
    """
    This class provides options to decorate a plot with more information.

    Parameters
    ----------
    title: string, optional (default='')
        title
    xlabel: string, optional (default='')
        the x axis label
    ylabel: string, optional (default='')
        the y axis label
    xlim: tuple, optional (default=(None, None))
        a tuple of min and max of x axis
    ylim: tuple, optional (default=(None, None))
        a tuple of min and max of y axis
    grid: boolean, optional (default=True)
        axes grids can be on (True) or off (False)
        check: https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.grid.html
    gridcolor: string, optional (default='k')
        set grid color
        check this link: https://matplotlib.org/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D.set_linestyle
    grid_linewidth: float, optional (default=2)
        set grid line width
        check this link: https://matplotlib.org/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D.set_linewidth

    Examples
    --------
    >>> from cheml.visualization import hist, decorator
    >>> from cheml.datasets import load_organic_density
    >>> smiles, density, features = load_organic_density()
    >>> hg = hist(20,'g',{'normed':True})
    >>> fig = hg.plot(features,'AMW')   # atomic molecular weights
    >>> dec = decorator('histogram',xlabel='density', ylabel='%', xlim=(4,None), ylim=(0,None),\
            grid=True, grid_color='g', grid_linestyle=':', grid_linewidth=0.5)
    >>> fig = dec.fit(fig)
    >>> fig.show()
    """
    def __init__(self, title='', xlabel='', ylabel='', xlim=(None, None), ylim=(None, None),
                 grid=True, grid_color='k', grid_linestyle='--', grid_linewidth=0.5):
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.xlim = xlim
        self.ylim = ylim
        self.grid = grid
        self.grid_color = grid_color
        self.grid_linestyle = grid_linestyle
        self.grid_linewidth = grid_linewidth

    def fit(self, figure):
        """
        the main function to fit the parameters to the input figure

        Parameters
        ----------
        figure: matplotlib.figure.Figure object or matplotlib.AxesSubplot object
            this function only proceed with one set of axes in the figure.

        Returns
        -------
        matplotlib.figure.Figure object

        """
        if str(type(figure)) == "<class 'matplotlib.axes._subplots.AxesSubplot'>":
            ax = figure
            figure = figure.figure
        elif str(type(figure)) == "<class 'matplotlib.figure.Figure'>":
            ax = figure.axes[0]
        else:
            msg = 'object must be a matplotlib.AxesSubplot or matplotlib.Figure object'
            raise TypeError(msg)
        if len(figure.axes) != 1:
            msg = 'matplotlib.figure object includes more than one axes'
            raise TypeError(msg)
        ax.set_title(self.title)
        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)
        ax.set_xlim(self.xlim[0], self.xlim[1])
        ax.set_ylim(self.ylim[0], self.ylim[1])
        ax.grid(color=self.grid_color, linestyle=self.grid_linestyle, linewidth=self.grid_linewidth)
        ax.grid(self.grid)
        return figure

    def matplotlib_font(self, family='normal', size=18, weight='normal', style='normal', variant='normal'):
        """
        The matplotlib_font function sets custom font properties.

        Notes
        -----
        Changing these parameters wil affect all the plots in your working session.

        Parameters
        ----------
        family: string, optional (default = 'normal')
            check this example: https://matplotlib.org/examples/pylab_examples/fonts_demo.html
        size: integer or string, optional (default = 18)
            check this example: https://matplotlib.org/examples/pylab_examples/fonts_demo.html
        weight: string, optional (default = 'normal')
            check this example: https://matplotlib.org/examples/pylab_examples/fonts_demo.html
        style: string, optional (default = 'normal')
            check this example: https://matplotlib.org/examples/pylab_examples/fonts_demo.html
        variant: string, optional (default = 'normal')
            check this example: https://matplotlib.org/examples/pylab_examples/fonts_demo.html

        Returns
        -------

        """
        matplotlib.rcParams.update(
            {'font.size': size, 'font.weight': weight, 'font.family': family, 'font.style': style,
             'font.variant': variant})

class scatter2D(object):
    """
    The scatter 2D plotting interface. It is built on top of the matplotlib.pyplot.plot function.

    Parameters
    ----------
    color: string, optional (default='b')
        set color
        check this link Notes for available options: https://matplotlib.org/api/_as_gen/matplotlib.pyplot.plot.html#matplotlib.pyplot.plot

    marker: string, optional (default='.')
        set marker
        check this link Notes for available options: https://matplotlib.org/api/_as_gen/matplotlib.pyplot.plot.html#matplotlib.pyplot.plot

    linestyle: string, optional (default='')
        set line style
        check this link: https://matplotlib.org/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D.set_linestyle
    linewidth: float, optional (default=2)
        set line width
        check this link: https://matplotlib.org/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D.set_linewidth

    Examples
    --------
    >>> from cheml.visualization import scatter2D
    >>> from cheml.datasets import load_organic_density
    >>> smiles, density, features = load_organic_density()
    >>> sc = scatter2D('r', marker='.')
    >>> fig = sc.plot(dfx=den, dfy=fea, x=0, y=1)
    >>> fig.show()
    """

    def __init__(self, color='b', marker='.', linestyle='', linewidth=2):
        self.color = color
        self.marker = marker
        self.linestyle = linestyle
        self.linewidth = linewidth

    def plot(self,dfx,dfy,x,y):
        """
        the main function to plot based on the input dataframes and their headers

        Parameters
        ----------
        dfx: pandas dataframe
            the x axis data
        dfy: pandas dataframe
            the y axis data
        x: string or integer, optional (default=0)
            header or position of data in the dfx
        y: string or integer, optional (default=0)
            header or position of data in the dfy

        Returns
        -------
        matplotlib.figure.Figure object

        """
        # check data
        if isinstance(x,str):
            X = dfx[x].values
        elif isinstance(x,int):
            X = dfx.iloc[:,x].values
        else:
            msg = 'x must be string for the header or integer for the postion of data in the dfx'
            raise TypeError(msg)
        
        if isinstance(y, str):
            Y = dfy[y].values
        elif isinstance(y, int):
            Y = dfy.iloc[:, y].values
        else:
            msg = 'y must be string for the header or integer for the postion of data in the dfy'
            raise TypeError(msg)

        # instantiate figure
        fig = plt.figure()
        ax = fig.add_subplot(111)
        trash = ax.plot(X,Y,color=self.color, marker=self.marker, linestyle=self.linestyle, linewidth= self.linewidth)
        return fig

class hist(object):
    """
    The histogram plotting interface. It is built on top of the matplotlib.pyplot.hist function .
    check this link: https://matplotlib.org/api/_as_gen/matplotlib.pyplot.hist.html#matplotlib.pyplot.hist

    Parameters
    ----------
    bins:  integer or sequence or 'auto', optional (default = None)
        exact same parameter from matplotlib.pyplot.hist

    color: color or array_like of colors or None, optional (default = None)
        exact same parameter from matplotlib.pyplot.hist

    kwargs : dictionary, optional (default = {})
        add any matplotlib.pyplot.hist parameter in the form of a dictionary.
        check this link: https://matplotlib.org/api/_as_gen/matplotlib.pyplot.hist.html#matplotlib.pyplot.hist

        provide keys in the form of a string.
        for example kwargs = {'key':value}

    Examples
    --------
    >>> from cheml.visualization import hist
    >>> from cheml.datasets import load_organic_density
    >>> smiles, density, features = load_organic_density()
    >>> hg = hist(20,'g',{'normed':True})
    >>> fig = hg.plot(features,'AMW')   # atomic molecular weights
    >>> fig.show()

    """
    def __init__(self,bins=None, color=None, kwargs={}):
        self.bins = bins
        self.color = color
        self.kwargs = kwargs

    def plot(self, dfx, x):
        """
        the main function to plot based on the input dataframe and its header

        Parameters
        ----------
        dfx: pandas dataframe
            the x axis data
        x: string or integer, optional (default=0)
            header or position of data in the dfx

        Returns
        -------
        matplotlib.figure.Figure object

        """
        # check data
        if isinstance(x, str):
            X = dfx[x].values
        elif isinstance(x, int):
            X = dfx.iloc[:, x].values
        else:
            msg = 'x must be string for the header or integer for the postion of data in the dfx'
            raise TypeError(msg)

        # instantiate figure
        fig = plt.figure()
        ax = fig.add_subplot(111)
        tash = ax.hist(X, bins= self.bins, color=self.color, **self.kwargs)
        return fig

class SavePlot(object):
    """
    Accepts a matplotlib AxesSubplot object and saves the figure with distinct options and at a specific location.
    Displays the path to the saved figure.

    Parameters:
    ----------
    filename: string
        name of the file that needs to be saved

    output_directory: string, optional (default=None)
        specify the folder where the figure needs to be saved.
        If the output directory that is specified does not exist, a new directory is created.

    format: string, optional (default='png')
        format of the figure that needs to be saved.
        Note: we recommend 'eps' for publication quality

    kwargs : dictionary, optional (default = {})
        add any matplotlib options in the form of a dictionary.
        provide keys in the form of a string.
        for example kwargs = {'key':value}
        https://matplotlib.org/api/_as_gen/matplotlib.figure.Figure.html#matplotlib.figure.Figure.savefig

    Example:
    --------
    >>> from cheml.datasets import load_cep_homo
    >>> from cheml.visualization import SavePlot
    >>> smiles, homo = load_cep_homo()
    >>> ax = homo.plot(kind='hist')
    >>> sa=SavePlot(ax,filename='homo',output_directory='plots',kwargs={'facecolor':'w','dpi':100,'pad_inches':0.1, 'bbox_inches':'tight'})
    >>> sa.save(main_directory='project')
    The Plot has been saved at:  project/plots/abc1.png

    """
    def __init__(self,filename, output_directory = None, format ='png',kwargs={}):
        self.filename = filename
        self.output_directory = output_directory
        self.format = format
        self.kwargs=kwargs

    def save(self, obj, main_directory='.'):
        """
        This is the main function that saves the figure.

        Parameters:
        ----------
        obj: matplotlib.axes._subplots.AxesSubplot or matplotlib.figure.Figure
            contains information about the plot

        main_directory: string, optional (default = '.')
            specify the parent directory where the folder needs to be saved.
            
        """
        if str(type(obj)) == "<class 'matplotlib.axes._subplots.AxesSubplot'>":
            obj = obj.figure
        elif str(type(obj)) == "<class 'matplotlib.figure.Figure'>":
            pass
        else:
            msg = 'object must be a matplotlib.AxesSubplot or matplotlib.Figure object'
            raise TypeError(msg)

        if self.output_directory:
            self.output_directory = os.path.join(main_directory, self.output_directory)
            if not os.path.exists(self.output_directory):
                os.makedirs(self.output_directory)
            self.file_path = '%s/%s.%s' % (self.output_directory, self.filename,self.format)

            obj.savefig(self.file_path,**self.kwargs)

        else:
            self.file_path = '%s/%s.%s' %(main_directory,self.filename,self.format)
            obj.figure.savefig(self.file_path,**self.kwargs)
        print 'The Plot has been saved at: ', self.file_path



# class Scatter_2D(object):
#     """
#     (Scatter_2D)
#     Uses matplotlib.pyplot to create a basic 2 dimensional plot with x data and y data.
#     (https://matplotlib.org/2.2.0/index.html)
#
#     Parameters:
#     ----------
#     xheader,yheader: string or list of strings, optional (default = ['x'],['y'])
#         string or list of strings containing the names of column headers of the dataframe that needs to be plot.
#
#     nod: integer, optional (default = 1)
#         number of data sets that need to be plot
#
#     dflist : pandas dataframe or list of pandas dataframes
#         specify the data that needs to be plot
#         provide this while calling the plot function
#
#     NOTE: x data and y data in the same dataframe ( specify xheaders and yheaders )
#         x data and y data in different
#         if dflist has 1 dataframe nod can be n; for example (dflist=[df1] ; nod = 1)
#         if dflist has n dataframes nod has to be n/2  (dflist=[df1,df2,df3,df4] ; nod=2)
#
#
#     legend: Boolean, optional (default=False)
#         True if a legend is required
#
#     xmin, xmax, ymin, ymax = integer, optional (default = 0)
#         set upper and lower limits of x and y axis in the plot
#
#     title: string, optional (default= 'Plot')
#         set the figure title
#
#     xlabel, ylabel : string, optional (default = 'x','y')
#         label x and y axes
#
#     sc : list of strings, optional (default = '')
#         mention the shape and color of the points on the plot, for example 'ro', 'b+', etc.
#
#     legend_title : list of strings, optional (default = [])
#         label each plot on the legend
#
#     l_pos : string, optional (default = 'Best')
#         specify the position of the legend.
#
#     kwargs : dictionary, optional (default = {})
#         add any matplotlib options in the form of a dictionary.
#         provide keys in the form of a string.
#         for example kwargs = {'key':value}
#
#
#
#     Example:
#     --------
#     >>> from cheml.datasets import load_organic_density
#     >>> smiles,density,features=load_organic_density()
#     >>> pl3=Scatter_2D(['MW'],['AMW'],nod=1,xlabel='MW',ylabel='AMW',sc=['bo'],title='MW vs AMW', kwargs={'markersize':0.5,'alpha':0.25},xmin=800,xmax=2000,ymin=800,ymax=2000)
#     >>> fig=pl3.plot(features)
#     >>> # fig.show()
#
#     """
#     def __init__(self, xheader=['x'], yheader=['y'],legend=False,xmin=0,xmax=0,ymin=0,ymax=0,xlabel= 'x', ylabel='y',title='Plot',sc=[''],legend_titles=[],l_pos='best',kwargs={}):
#
#         self.l_pos = l_pos
#         self.legend=legend
#         self.legend_titles=legend_titles
#         self.sc=sc
#         self.xmin=xmin
#         self.xmax=xmax
#         self.ymin=ymin
#         self.ymax=ymax
#         self.xlabel = xlabel
#         self.ylabel = ylabel
#         self.title = title
#         self.kwargs=kwargs
#         self.xheader=xheader
#         self.yheader=yheader
#
#     def __fit(self,i,j):
#
#         if j==1:
#             if len(self.dfx)==1:
#                 dfx=self.dfx
#                 for i in range(len(self.dfy)-1):
#                     dfx.append(self.dfx[0])
#                 self.dfx=dfx
#             elif len(self.dfy)==1:
#                 dfy=self.dfy
#                 for i in range(len(self.dfx)-1):
#                     dfy.append(self.dfy[0])
#                 self.dfy=dfy
#             else:
#                 msg="length of dataframe lists (dfx and dfy) should be same, or one of them should be equal to 1"
#                 raise IOError(msg)
#
#         elif j==2:
#             if len(self.dfx)==1:
#                 dfx=self.dfx
#                 for i in range(len(self.xheader)-1):
#                     dfx.append(self.dfx[0])
#                 self.dfx=dfx
#             elif len(self.xheader)==1:
#                 xheader=self.xheader
#                 for i in range(len(self.dfx)-1):
#                     xheader.append(self.xheader[0])
#                 self.xheader = xheader
#             else:
#                 msg = "length of header list (xheader) should be same as dataframe list (dfx) or one of them should be equal to 1"
#                 raise IOError(msg)
#
#         elif j==3:
#             if len(self.dfy)==1:
#                 dfy=self.dfy
#                 print 'here'
#                 for i in range(len(self.yheader)-1):
#                     dfy.append(self.dfy[0])
#                 self.dfy=dfy
#             elif len(self.yheader)==1:
#                 yheader=self.yheader
#                 for i in range(len(self.dfy)-1):
#                     yheader.append(self.yheader[0])
#                 self.yheader = yheader
#             else:
#                 msg = "length of header list (yheader) should be same as dataframe list (dfy) or one of them should be equal to 1"
#                 raise IOError(msg)
#         elif j==4:
#             if len(self.xheader)==1:
#                 xheader=self.xheader
#                 for i in range(len(self.yheader)-1):
#                     xheader.append(self.xheader[0])
#                 self.xheader=xheader
#             elif len(self.yheader)==1:
#                 yheader=self.yheader
#                 for i in range(len(self.xheader)-1):
#                     yheader.append(self.yheader[0])
#                 self.yheader=yheader
#             else:
#                 msg="length of header lists (xheader and yheader) should be the same or one of them should be equal to 1"
#                 raise IOError(msg)
#
#         else:
#             dfx=self.dfx[i]
#             dfy=self.dfy[i]
#             if isinstance(self.xheader[i], str):
#                 self.x = dfx[self.xheader[i]]
#             else:
#                 self.x = dfx.iloc[:, self.xheader[i]]
#
#             if isinstance(self.yheader[i],str):
#                 self.y = dfy[self.yheader[i]]
#             else:
#                 self.y = dfy.iloc[:, self.yheader[i]]
#
#     def plot(self,dfx,dfy):
#         """
#         (plot)
#         This is the main function to create a 2D plot from dflist
#
#         Parameters:
#         ----------
#         dflist : pandas dataframe or list of pandas dataframes
#             specify the data that needs to be plot.
#             provide this while calling the plot function
#
#         Returns:
#         ------
#         f : matplotlib object
#             returns the matplotlib object containing the plot information
#
#         """
#
#         self.dfx=dfx
#         self.dfy=dfy
#
#         if len(self.xheader)!=len(self.dfx):
#             self.__fit(-1,2)
#
#         if len(self.yheader)!=len(self.dfy):
#             self.__fit(-1,3)
#
#         if len(self.dfx)!=len(self.dfy):
#             self.__fit(-1, 1)
#
#         if len(self.xheader)!=len(self.yheader):
#             self.__fit(-1,4)
#
#         if len(self.dfx)==1:
#             f=plt.figure()
#             self.__fit(0,0)
#             if self.sc==['']:
#                 plt.plot(self.x,self.y,**self.kwargs)
#             else:
#                 plt.plot(self.x,self.y,self.sc[0],**self.kwargs)
#
#         else:
#             f, ax = plt.subplots()
#             for i in range(len(self.dfx)):
#                 self.__fit(i,0)
#                 x1=self.x
#                 y1=self.y
#                 if self.sc[i]!='':
#                     plt.plot(x1,y1,self.sc[i],**self.kwargs)
#                 else:
#                     plt.plot(x1,y1,**self.kwargs)
#
#         plt.xlabel(self.xlabel)
#         plt.ylabel(self.ylabel)
#         plt.title(self.title)
#
#         if self.xmin+self.xmax!=0 and self.ymin+self.ymax!=0:
#             plt.axis([self.xmin,self.xmax,self.ymin,self.ymax])
#         if self.legend==True:
#             plt.legend(self.legend_titles,loc=self.l_pos)
#         return f

# class hist(object):
#     """
#     (hist)
#     Uses matplotlib.pyplot to create a simple histogram.
#     (https://matplotlib.org/2.2.0/index.html)
#
#     Parameters:
#     ----------
#     nbins: integer
#         number of bins
#
#     rwidth: integer, optional (default=1)
#         width of the bars in the histogram
#
#     xmin,xmax: integer, optional(default = 0,0)
#         set upper and lower limits of the x axis
#
#     bestfit: Boolean, optional (default = False)
#
#     linshapecolor: string, optional (default = '')
#         shape and color of the lines in the histogram
#
#     xlabel, ylabel: strings, optional (default = 'x','y')
#         label for the x and y axes
#
#     title : string, optional (default = 'histogram')
#         title for the figure
#
#     isformatter: Boolean, optional (default= False)
#
#     formatter_type: string, optional (default = '')
#
#     kwargs: dictionary, optional (default ={})
#         add any matplotlib options in the form of a dictionary.
#         provide keys in the form of a string.
#         for example kwargs = {'key':value}
#
#     Example:
#     --------
#     >>> from cheml.datasets import load_organic_density
#     >>> smiles,density,features=load_organic_density()
#     >>> from cheml.visualization import hist
#     >>> pl4=hist(nbins= 20, kwargs={'normed':True,'facecolor':'blue', 'ec':'black'}, lineshapecolor='g-', formatter_type='percent', rwidth=0.8, xmin=800, xmax=2000)
#     >>> fig1=pl4.plot(density)
#
#     """
#     def __init__(self,nbins,rwidth=1,xmin=0,xmax=0,bestfit=False,lineshapecolor='',xlabel='x',ylabel='y',title='histogram',isformatter=False,formatter_type='',kwargs={}):
#         self.nbins=nbins
#         self.rwidth=rwidth
#         self.xmin=xmin
#         self.lineshapecolor=lineshapecolor
#         self.bestfit=bestfit
#         self.xmax=xmax
#         self.title=title
#         self.isformatter=isformatter
#         self.formatter_type=formatter_type
#         self.xlabel=xlabel
#         self.ylabel=ylabel
#         self.kwargs=kwargs
#         self.mean=[]
#         self.std=[]
#
#     def __fit(self,df):
#         mean=[]
#         std=[]
#         self.x=np.asarray(df)
#         df.mean()
#         dfnew=df.describe()
#         mean.append(dfnew.loc['mean'])
#         std.append(dfnew.loc['std'])
#         self.mean=mean
#         self.std=std
#
#         print self.mean
#         print self.std
#
#
#     def plot(self,dflist):
#         """
#         (plot)
#         This is the main function to create a simple histogram from dflist
#
#         Parameters:
#         ----------
#         dflist : pandas dataframe or list of pandas dataframes
#             specify the data that needs to be plot.
#             provide this while calling the plot function
#
#         Returns:
#         ------
#         f : matplotlib object
#             returns the matplotlib object containing the plot information
#
#         """
#         if not isinstance(dflist, list)==True:
#             self.__fit(dflist)
#             self.dflist=[dflist]
#         else:
#             for i in dflist:
#                 self.__fit(i)
#             self.dflist=dflist
#
#         f,ax=plt.subplots()
#
#         if self.isformatter==True:
#             if self.formatter_type=='percent':
#                 formatter=matplotlib.ticker.EngFormatter()
#                 ax.yaxis.set_major_formatter(formatter)
#
#         if (self.xmin+self.xmax)!=0:
#             n,bins, patches=plt.hist(self.x,bins=self.nbins,rwidth=self.rwidth,range=(self.xmin,self.xmax),**self.kwargs)
#         else:
#             n,bins, patches=plt.hist(self.x,bins=self.nbins,rwidth=self.rwidth,**self.kwargs)
#
#
#         if self.bestfit==True and self.lineshapecolor!='':
#             y = mlab.normpdf(bins)#, self.mean, self.std)
#             plt.plot(bins,y,self.lineshapecolor,**self.kwargs)
#
#         plt.xlabel(self.xlabel)
#         plt.ylabel(self.ylabel)
#         plt.title(self.title)
#         return f



"""
The cheml.search module includes (please click on links adjacent to function names for more information):
    - GA_DEAP: :func:`~cheml.search.GA_DEAP`
"""

from .genetic_algorithm import GA_DEAP

__all__ = [
    'GA_DEAP',
]



import unittest
import numpy as np
import sys
from .....models.regression.crystal.CoulombSineMatrixRegression import \
    CoulombSineMatrixRegression
from .....vassal.data.Atom import Atom
from .....vassal.data.Cell import Cell

class testCoulombSineMatrixRegression(unittest.TestCase):
    def setUp(self):
        self.r = CoulombSineMatrixRegression()

    def tearDown(self):
        self.r = None

    def test_matrix(self):
        # Make a simple structure.
        structure = Cell()
        structure.add_atom(Atom([0, 0, 0], 0))
        structure.set_type_name(0, "Al")

        # Compute the sine matrix.
        mat = self.r.compute_coulomb_matrix(structure)
        self.assertEquals(1, mat.shape[0])
        self.assertEquals(1, mat.shape[1])
        self.assertAlmostEquals(0.5 * 13 ** 2.4, mat[0, 0],delta=1e-6)

        # Add another atom and repeat.
        structure.add_atom(Atom([0.5, 0.5, 0.5], 0))
        mat = self.r.compute_coulomb_matrix(structure)
        self.assertEquals(2, mat.shape[0])
        self.assertEquals(2, mat.shape[1])

        # Test: Is it insensitive to basis changes.
        new_basis = structure.get_basis()
        new_basis[1, 0] = 12
        structure.set_basis(basis=new_basis)
        self.assertAlmostEquals(1.0, structure.volume(), delta=1e-6)
        mat2 = self.r.compute_coulomb_matrix(structure)
        if np.linalg.norm(mat - mat2) > 1e-6:
            sys.stderr.write("WARNING: Not insensitive to basis changes\n")

    def test_representation(self):
        # Make a simple structure.
        structure = Cell()
        structure.add_atom(Atom([0, 0, 0], 0))
        structure.set_type_name(0, "Al")

        # Compute the sine matrix.
        mat = self.r.compute_representation(structure)
        self.assertEquals(1, len(mat))

        # Add another atom and repeat.
        structure.add_atom(Atom([0.5, 0.5, 0.5], 0))
        mat = self.r.compute_coulomb_matrix(structure)
        self.assertEquals(2, len(mat))

        # Test: Is it insensitive to basis changes.
        new_basis = structure.get_basis()
        new_basis[1, 0] = 12
        structure.set_basis(basis=new_basis)
        self.assertAlmostEquals(1.0, structure.volume(), delta=1e-6)
        mat2 = self.r.compute_representation(structure)
        if np.sum(mat - mat2) / len(mat) > 1e-6:
            sys.stderr.write("WARNING: Not insensitive to basis changes\n")

    def test_distance(self):
        # Make two simple structures.
        structure1 = Cell()
        structure1.add_atom(Atom([0, 0, 0], 0))
        structure1.set_type_name(0, "Al")

        structure2 = Cell()
        structure2.add_atom(Atom([0, 0, 0], 0))
        structure2.set_type_name(0, "Ni")

        # Compute representations of each structure.
        rep1 = self.r.compute_representation(structure1)
        rep2 = self.r.compute_representation(structure2)

        # Check that similarity between identical structures is 1.0.
        self.assertAlmostEquals(1.0, self.r.compute_similarity(rep1, rep1),
                                delta=1e-6)
        self.assertAlmostEquals(1.0, self.r.compute_similarity(rep2, rep2),
                                delta=1e-6)

        # Check symmetry.
        self.assertAlmostEquals(self.r.compute_similarity(rep1, rep2),
                                self.r.compute_similarity(rep2, rep1),
                                delta=1e-6)

        # Check that similarity between these structures is less than 1.0.
        self.assertTrue(self.r.compute_similarity(rep1, rep2) < 1.0)


import numpy as np
import math
from ....data.materials.util.LookUpData import LookUpData

class CoulombSineMatrixRegression:
    """Class to perform regression based on the Coulomb Sine Matrix approach of
    Faber et al. [1].

    LW 1Apr15: This method does not appear to be insensitive to basis cell
    selection.

    Attributes
    ----------
    sigma : float
        Normalization term in kernel function.

    References
    ----------
    .. [1] F. Faber, A. Lindmaa, O. A. von Lilienfeld, and R. Armiento,
    "Crystal structure representations for machine learning models of
    formation energies," International Journal of Quantum Chemistry,
    vol. 115, no. 16, pp. 1094--1101, Apr. 2015.

    """

    def __init__(self):
        """Function to create instance and initialize fields.
        """

        # Normalization term in kernel function.
        self.sigma = 1

    def set_sigma(self, s):
        """Function to set the normalization parameter in the kernel function.

        Parameters
        ----------
        s : float
            Desired normalization parameter.
        """
        self.sigma = s

    def compute_similarity(self, structure1, structure2):
        """Function to compute similarity between two crystal structures.

        Parameters
        ----------
        structure1 : array-like
            Representation of structure #1.
        structure2 : array-like
            Representation of structure #2.

        Returns
        -------
        output : float
            Similarity between the two structures.

        """
        e1 = structure1.copy()
        e2 = structure2.copy()

        # Determine which is bigger.
        flag = len(e1) > len(e2)
        len_diff = len(e1) - len(e2)
        if flag:
            e2 = np.lib.pad(e2, (0, len_diff), 'constant', constant_values=(0))
            diff = e1 - e2
        else:
            e1 = np.lib.pad(e1, (0, -len_diff), 'constant', constant_values=(
                0))
            diff = e2 - e1

        # Compute the L1 distance.
        dist = sum(np.abs(diff))

        # Compute the Laplacian.
        return math.exp(-1 * dist / self.sigma)

    def compute_representation(self, structure):
        """Function to compute the representation of the crystal structure.

        Parameters
        ----------
        structure : CrystalStructureEntry
            Crystal structure.

        Returns
        -------
        output : array-like
            Representation of the structure.

        """

        # First generate the Coulomb matrix.
        matrix = self.compute_coulomb_matrix(structure)

        # Compute the eigenvalues.
        w, v = np.linalg.eig(matrix)
        return w

    def compute_coulomb_matrix(self, structure):
        """Function to compute the Coulomb sine matrix. Equation 24 of the
        paper describing this method.

        Parameters
        ----------
        structure : CrystalStructureEntry
            Structure to be evaluated.

        Returns
        -------
        output : array-like
            Coulomb sine matrix.

        Raises
        ------
        Exception
            If element is not found.
        """

        # Get basis vectors.
        basis = structure.get_basis()

        # Get reciprocal basis vectors.
        basis_inverse = structure.get_inverse_basis()

        # Create output matrix.
        n_atoms = structure.n_atoms()
        n_types = structure.n_types()
        output = np.zeros((n_atoms, n_atoms), dtype=float)

        # Get the positions for each atom.
        pos = np.array([structure.get_atom(a).get_position_cartesian() for a
                        in range(n_atoms)])

        # Get the atomic number of each element.
        type_z = np.zeros(n_types, dtype=int)
        for i in range(n_types):
            if structure.get_type_name(i) in LookUpData.element_names:
                type_z[i] = 1 + LookUpData.element_names.index(
                            structure.get_type_name(i))
            else:
                raise Exception("No such element: "+structure.get_type_name(i))

        # Compute all terms.
        for r1 in range(n_atoms):
            output[r1, r1]  = 0.5 * (type_z[structure.get_atom(r1).get_type(
                                )]) ** 2.4
            for r2 in range(r1 + 1, n_atoms):
                # Displacement between the two atoms.
                disp = pos[r1] - pos[r2]

                # Convert to fractional coordinates.
                disp = np.matmul(basis_inverse, disp)

                # Multiply by pi and compute sin^2 of each element.
                disp = np.sin(disp * np.math.pi) ** 2

                # Multiply by basis vectors.
                disp = np.matmul(basis, disp)

                # Get the result.
                res = 1.0 / np.linalg.norm(disp)
                res *= type_z[structure.get_atom(r1).get_type()] * type_z[
                    structure.get_atom(r2).get_type()]
                output[r1, r2] = output[r2, r1] = res

        return output


import pandas as pd
import numpy as np
import types
from ....data.materials.CrystalStructureEntry import CrystalStructureEntry

class PackingEfficiencyAttributeGenerator:
    """Class to compute attributes based on packing efficiency.

    Packing efficiency is determined by finding the largest sphere that would
    fit inside each Voronoi cell and comparing the volume of that sphere to the
    volume of the cell.

    Notes
    -----
    For now, the only attribute computed by this generator is the maximum
    packing efficiency for the entire cell. This is computed by summing the
    total volume of all spheres in all cells, and dividing by the volume of
    the unit cell.

    """

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Crystal structures for which features are to be generated. A list
            of CrystalStructureEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CrystalStructureEntry instances.

        """

        # Initialize list of feature values for pandas data frame.
        feat_values = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")
        elif (entries and not isinstance(entries[0], CrystalStructureEntry)):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")

        # Insert header names here.
        feat_headers = ["MaxPackingEfficiency"]

        # Generate features for each entry.
        for entry in entries:
            tmp_list = []
            # Get the Voronoi tessellation.
            try:
                voro = entry.compute_voronoi_tessellation()
            except Exception:
                tmp_list = [np.nan] # If tessellation fails.
                feat_values.append(tmp_list)
                continue
            tmp_list.append(voro.max_packing_efficiency())
            feat_values.append(tmp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import unittest
import numpy as np
from cheml.chem import CoulombMatrix

# Oxygen, Hydrogen, Hydrogen
n = np.array([8, 1, 1])
c = np.array([[1.464, 0.707, 1.056],
             [0.878, 1.218, 0.498],
             [2.319, 1.126, 0.952]])
nc = n.reshape((3,1))
nc = np.append(nc,c,1)

class TestCoulombMatrix(unittest.TestCase):
    def test_UM(self):
        cm = CoulombMatrix('UM')
        h2o = cm.represent(np.array([nc]))

        self.assertEqual(h2o.shape, (1, cm.max_n_atoms**2))
        a = np.array([[ 73.51669472, 8.3593106 ,8.35237809, 8.3593106,
          0.5, 0.66066557, 8.35237809, 0.66066557, 0.5]])
        self.assertAlmostEqual(a[0][0], h2o.values[0][0])
        self.assertAlmostEqual(a[0][1], h2o.values[0][1])
        self.assertAlmostEqual(a[0][-1], h2o.values[0][-1])

    def test_UT(self):
        cm = CoulombMatrix('UT')
        h2o = cm.represent(np.array([nc]))

        self.assertEqual(h2o.shape, (1, cm.max_n_atoms*(cm.max_n_atoms+1)/2))
        a = np.array([[ 73.51669472, 8.3593106, 0.5, 8.35237809,
          0.66066557, 0.5]])
        self.assertAlmostEqual(a[0][0], h2o.values[0][0])
        self.assertAlmostEqual(a[0][1], h2o.values[0][1])
        self.assertAlmostEqual(a[0][-1], h2o.values[0][-1])

    def test_E(self):
        cm = CoulombMatrix('E')
        h2o = cm.represent(np.array([nc]))

        self.assertEqual(h2o.shape, (1, cm.max_n_atoms))
        a = np.array([[ 75.39770052,  -0.16066482,  -0.72034098]])
        self.assertAlmostEqual(a[0][0], h2o.values[0][0])
        self.assertAlmostEqual(a[0][1], h2o.values[0][1])
        self.assertAlmostEqual(a[0][-1], h2o.values[0][-1])


    def test_SC(self):
        cm = CoulombMatrix('SC')
        h2o = cm.represent(np.array([nc]))

        self.assertEqual(h2o.shape, (1, cm.max_n_atoms*(cm.max_n_atoms+1)/2))
        a = np.array([[ 73.51669472,   8.3593106 ,   0.5       ,   8.35237809,
          0.66066557,   0.5       ]])
        self.assertAlmostEqual(a[0][0], h2o.values[0][0])
        self.assertAlmostEqual(a[0][1], h2o.values[0][1])
        self.assertAlmostEqual(a[0][-1], h2o.values[0][-1])

    def test_RC(self):
        cm = CoulombMatrix('RC')
        h2o = cm.represent(np.array([nc]))

        self.assertEqual(h2o.shape, (1, cm.nPerm*cm.max_n_atoms*(cm.max_n_atoms+1)/2))
        a = np.array([[  0.5,   8.35237809,  73.51669472,   0.66066557,
          8.3593106 ,   0.5       ,  73.51669472,   8.35237809,
          0.5       ,   8.3593106 ,   0.66066557,   0.5       ,
          0.5       ,   8.3593106 ,  73.51669472,   0.66066557,
          8.35237809,   0.5       ]])




if __name__== '__main__':
    unittest.main()




import numpy as np
from numpy.linalg import norm
from .AtomImage import AtomImage

class Cell:
    """Represents the volume enclosing a set of atoms.

    Could represent a crystal unit cell or the simulation cell from a
    molecular dynamics calculation.

    Attributes
    ----------
    simulation_cell : array-like
        Vectors representing the sides of cell. Stored in the format:
        # a_x b_x c_x
        # a_y b_y c_y
        # a_z b_z c_z
    inverse_cell : array-like
        Inverse of the simulation cell. Set during the set_basis function.
    atoms : list
        List of all atoms.
    face_is_periodic : array-like
        Whether the walls of the cell are periodic.
    type_name : list
        Names of each atom type.
    lattice_vectors : array-like
        Lattice vectors.
    recip_lattice_vectors : array-like
        Reciprocal lattice vectors.
    """

    def __init__(self):
        """Creates an empty structure a = b = c = 1, alpha = beta = gamma = 90,
        and periodic boundary conditions in all directions.
        """

        # Vectors representing the sides of cell. Stored in the format:
        # a_x b_x c_x
        # a_y b_y c_y
        # a_z b_z c_z
        self.simulation_cell = None

        # Inverse of the simulation cell. Set during the set_basis function.
        self.inverse_cell = None

        # List of all atoms.
        self.atoms = []

        # Whether the walls of the cell are periodic.
        self.face_is_periodic = None

        # Names of each atom type.
        self.type_name = []

        # Lattice vectors.
        self.lattice_vectors = None

        # Reciprocal lattice vectors.
        self.recip_lattice_vectors = None

        basis = np.eye(3, dtype=float)
        self.simulation_cell = np.eye(3, dtype=float)
        self.set_basis(basis=basis)
        self.face_is_periodic = np.array([True]*3)

    def __copy__(self):
        """Function to override the copy() method.

        Returns
        -------
        output : Cell
            A new instance with the appropriate properties set.
        """
        x = type(self)(self.__class__)
        x.__dict__.update(self.__dict__)
        x.atoms = []
        for atom in self.atoms:
            n_atom = atom.__copy__()
            n_atom.set_cell(x)
            x.atoms.append(n_atom)

        x.face_is_periodic = self.face_is_periodic.copy()
        x.inverse_cell = self.inverse_cell.copy()
        x.simulation_cell = self.simulation_cell.copy()
        x.lattice_vectors = self.lattice_vectors.copy()
        x.recip_lattice_vectors = self.recip_lattice_vectors.copy()
        x.type_name = list(self.type_name)
        return x


    def __eq__(self, other):
        """Function to override the check for equality with another object of
        the same class.

        Parameters
        ----------
        other : Cell
            Other object.

        Returns
        -------
        output : bool
            True if equal, else False.
        """

        if isinstance(other, Cell):
            # Check to see if number of atoms is the same.
            if (self.n_atoms() != other.n_atoms()):
                return False
            # Compare basis.
            diff = self.simulation_cell - other.simulation_cell
            if max([sum(abs(diff[:, i])) for i in range(3)]) > 1e-4:
                return False

            # Compare atoms.
            for atom1 in self.atoms:
                # See if any atom in other matches this one regardless of ID.
                found = False
                for atom2 in other.atoms:
                    if atom1.__eq__(atom2):
                        found = True
                        break
                if not found:
                    return False

            return True
        return False

    def set_basis(self, basis=None, lengths=None, angles=None):
        """Function to define the basis.

        Format:
        a_x b_x c_x
        a_y b_y c_y
        a_z b_z c_z

        Parameters
        ----------
        basis : array-like
            2-D numpy array containing the new basis. (Default value = None)
        lengths : array-like
            List of lengths that define the basis. (Default value = None)
        angles : array-like
            List of angles that define the basis. (Default value = None)

        Raises
        ------
        ValueError
            If input doesn't contain 3 lengths and angles.
            If angles are not between 0 and 180.
            If neither basis nor the lengths and angles are specified.
            If matrix is not 3x3.
            If provided basis has non-positive volume.
        """

        c_basis = basis

        if c_basis is None and lengths is not None and angles is not None:
            # Check input.
            if len(lengths) != 3 and len(angles) != 3:
                raise ValueError("Expected 3 lengths and/or 3 angles.")

            # Convert angles to radians.
            a = np.array(angles, dtype=float)
            if (a < 0).any() or (a > 180).any():
                raise ValueError("Angles must be between 0 and 180. Angle #")
            angles_radians = np.array(map(np.math.radians, a), dtype=float)
            l = np.array(lengths, dtype=float)
            c_basis = self.compute_basis(l, angles_radians)

        if c_basis is None:
            raise ValueError("Either basis must be specified or the lengths "
                             "and angles.")
        # Check input.
        if c_basis.shape != (3, 3):
            raise ValueError("Expected 3x3 matrix.")

        # Store old basis.
        old_cell = self.simulation_cell.copy()
        self.simulation_cell = c_basis

        # Make sure it has positive volume.
        if self.volume() <= 0 or np.isinf(self.volume()):
            self.simulation_cell = old_cell
            raise ValueError("Provided basis has non-positive volume.")

        # Get the inverse.
        self.inverse_cell = np.linalg.inv(self.simulation_cell)

        # Update cartesian coordinates of each atom.
        if self.atoms:
            for atom in self.atoms:
                atom.update_cartesian_coordinates()

        # Precompute the lattice vectors.
        self.lattice_vectors = np.zeros((3, 3), dtype=float)
        self.recip_lattice_vectors = np.zeros((3, 3), dtype=float)

        for i in range(3):
            self.lattice_vectors[i, :] = self.simulation_cell[:, i]
            self.recip_lattice_vectors[i, :] = self.inverse_cell[:, i]

    def compute_basis(self, lengths, angles_radians):
        """Function to compute the basis given lengths and angles (in radians).

        The a lattice vectors will be aligned in x direction, and the be will
        be in the xy plane.

        Parameters
        ----------
        lengths : array-like
            Lengths of the lattice vectors.
        angles_radians : array-like
            Lattice angles in radians.

        Returns
        -------
        output : array-like
            One possible set of basis vectors. Format:
            a_x b_x c_x
            a_y b_y c_y
            a_z b_z c_z

        """

        # Convert lengths to basis.
        basis = np.zeros((3, 3), dtype=float)
        basis[0][0] = lengths[0]
        basis[0][1] = lengths[1] * np.math.cos(angles_radians[2])
        basis[0][2] = lengths[2] * np.math.cos(angles_radians[1])
        basis[1][1] = lengths[1] * np.math.sin(angles_radians[2])
        basis[1][2] = lengths[2] * (np.math.cos(angles_radians[0]) -
                    np.math.cos(angles_radians[1]) * np.math.cos(
                    angles_radians[2]))/ np.math.sin(angles_radians[2])

        v = np.math.sqrt(1 - np.math.cos(angles_radians[0]) * np.math.cos(
            angles_radians[0]) - np.math.cos(angles_radians[1]) * np.math.cos(
            angles_radians[1]) - np.math.cos(angles_radians[2]) * np.math.cos(
            angles_radians[2]) + 2 * np.math.cos(angles_radians[0]) *
            np.math.cos(angles_radians[1]) * np.math.cos(angles_radians[2]))
        basis[2][2] = lengths[2] * v / np.math.sin(angles_radians[2])
        return basis

    def add_atom(self, a):
        """Function to add atom to cell.

        Parameters
        ----------
        a : Atom
            Atom to be added.

        """
        self.atoms.append(a)
        a.set_id(len(self.atoms) - 1)
        a.set_cell(self)
        while len(self.type_name) <= a.get_type():
            self.type_name.append(None)

    def direction_is_periodic(self, index):
        """Function to return whether a certain direction has boundary
        conditions.

        Parameters
        ----------
        index : int
            Desired direction (0: x, 1: y, 2: z)

        Returns
        -------
        output : bool
            True if that direction has periodic boundary conditions,
            else False.

        """
        return self.face_is_periodic[index]

    def volume(self):
        """Function to compute the volume of the simulation cell.

        Returns
        -------
        output : float
            Volume of the cell.
        """
        return self.simulation_cell[0][0] * (self.simulation_cell[1][1] *
                self.simulation_cell[2][2] - self.simulation_cell[2][1] *
                self.simulation_cell[1][2])- self.simulation_cell[0][1] * (
                self.simulation_cell[1][0] * self.simulation_cell[2][2] -
                self.simulation_cell[1][2] * self.simulation_cell[2][0]) + \
                self.simulation_cell[0][2] * (self.simulation_cell[1][0] *
                self.simulation_cell[2][1] - self.simulation_cell[1][1] *
                self.simulation_cell[2][0])

    def get_basis(self):
        """Function to get the basis of this structure.

        Returns
        -------
        output : array-like
            A copy of the 2-D numpy array that defines the basis.
        """
        return self.simulation_cell.copy()

    def get_basis_matrix(self):
        """Function to get the basis of this structure.

        Returns
        -------
        output : array-like
            A numpy matrix representing the basis.

        """
        return np.matrix(self.simulation_cell)

    def get_inverse_basis(self):
        """Function to get the inverse basis.

        Returns
        -------
        output : array-like
            A copy of the 2-D numpy array that defines the inverse basis.
        """
        return self.inverse_cell.copy()

    def get_lattice_vectors(self):
        """Function to get the lattice vectors for this cell.

        Returns
        -------
        output : array-like
            A 2-D numpy array where each row contains a lattice vector
            (0: a, 1: b, 2: c).
        """
        return self.lattice_vectors

    def get_lattice_parameters(self):
        """Function to get the lattice parameters.

        Returns
        -------
        output : array-like
            A numpy array containing the lattice parameters.
        """

        output = np.array(map(norm, [x for x in self.simulation_cell.T]),
                          dtype=float)
        return output

    def get_lattice_angles_radians(self, radians=True):
        """Function to get the angles between the lattice vectors.

        Parameters
        ----------
        radians : bool
             (Default value = True)

        Returns
        -------
        output : array-like
            Lattice angles in radians.
        """

        output_radians = []
        col0 = self.simulation_cell[:, 0]
        col1 = self.simulation_cell[:, 1]
        col2 = self.simulation_cell[:, 2]
        nc0 = norm(col0)
        nc1 = norm(col1)
        nc2 = norm(col2)
        # Compute cosines.
        output_radians.append(np.math.acos(col1.dot(col2) / (nc1 * nc2)))
        output_radians.append(np.math.acos(col2.dot(col0) / (nc2 * nc0)))
        output_radians.append(np.math.acos(col0.dot(col1) / (nc0 * nc1)))
        if radians:
            return np.array(output_radians)
        output_degrees = np.array(map(np.math.degrees, output_radians),
                                  dtype=float)
        return output_degrees

    def get_aligned_basis(self):
        """Function to get the basis vectors aligned such that a vector is
        parallel to the x-axis, and the b vector is in the x-y plane.

        Format:
        a_x b_x c_x
        a_y b_y c_y
        a_z b_z c_z


        Returns
        -------
        output : array-like
            Aligned basis vectors as 2-D numpy array.
        """

        angles = self.get_lattice_angles_radians()
        lengths = self.get_lattice_parameters()
        return self.compute_basis(lengths, angles)

    def get_reciprocal_vectors(self):
        """Function to get the reciprocal lattice vectors for this cell.

        These are simply the matrix inverse of the lattice vectors.

        Returns
        -------
        output : array-like
            A 2-D numpy array where each row contains a reciprocal
            vector (0: a, 1: b, 2: c).
        """
        return self.recip_lattice_vectors

    def get_atoms(self):
        """Function to get all atoms in the structure.

        Returns
        -------
        output : array-like
            A list of all atoms.
        """
        return self.atoms

    def get_atom(self, index):
        """Function to get a single atom.

        Parameters
        ----------
        index : int
            Index of atom.

        Returns
        -------
        output : Atom
            Atom at the desired index.
        """
        return self.atoms[index]

    def n_atoms(self):
        """Function to get the number of atoms in simulation cell.

        Returns
        -------
        output : int
            Number of atoms.
        """
        return len(self.atoms)

    def n_types(self):
        """Function to get the number of atom types.

        Returns
        -------
        output : int
            Number of atom types.

        """
        return len(self.type_name)

    def number_of_type(self, type):
        """Function to get the number of atoms of a certain type.

        Parameters
        ----------
        type : int
            Type index.

        Returns
        -------
        output : int
            Number of atoms.

        """
        count = 0
        for atom in self.atoms:
            if atom.get_type() == type:
                count += 1

        return count

    def add_type(self, name=None):
        """Function to add a new atom type to this cell.

        Parameters
        ----------
        name : str
            Name of type (can be None) (Default value = None)


        """
        self.type_name.append(name)

    def set_type_name(self, index, name):
        """Function to define the name for an atom type.

        Note that indexing starts at 0.

        Parameters
        ----------
        index : int
            Index of atom type.
        name : str
            Desired name.

        Raises
        ------
        IndexError
            If index is out of bounds.
        """

        if index < 0 or index > len(self.type_name):
            raise IndexError("Index out of bounds: "+index)
        self.type_name[index] = name

    def replace_type_names(self, changes):
        """Function to change the names of several types at once.

        Parameters
        ----------
        changes : dict
            Dictionary containing the changes to be made. Key:
            Current Name, Value: New Name
        """

        # Get the type ids corresponding to each name.
        name_map = {}
        for name in changes:
            name_map[name] = []

        for id, name in enumerate(self.type_name):
            for by_name in name_map:
                if by_name == name:
                    name_map[name].append(id)

        # Make the replacements.
        for by_name in name_map:
            new_name = changes[by_name]
            if not new_name:
                continue
            for id in name_map[by_name]:
                self.type_name[id] = new_name

    def merge_like_types(self):
        """Function to combine types that have the same name.

        """

        cur_type = 0
        while cur_type < self.n_types():
            cur_name = self.get_type_name(cur_type)
            if cur_name is None:
                # Skip if name is None.
                cur_type += 1
                continue

            # Look to see if there is a duplicate name.
            matched_type = -1
            for t in range(self.n_types()):
                if t == cur_type:
                    continue
                name = self.get_type_name(t)
                if name is None:
                    continue
                if name == cur_name:
                    matched_type = t
                    break

            # If not matched, increment and move on.
            if matched_type == -1:
                cur_type += 1
                continue

            # Merge (matched_type -> cur_type) (a -> matched_type -> a-1)
            self.type_name.remove(self.type_name[matched_type])
            for atom in self.atoms:
                type = atom.get_type()
                if type == matched_type:
                    atom.set_type(cur_type)
                elif type > matched_type:
                    atom.set_type(type - 1)

    def set_type_radius(self, index, radius):
        """Function to set the radius of all atoms of a certain type.

        Parameters
        ----------
        index : int
            Index of the desired type.
        radius : float
            Desired radius value.

        """

        self.atoms[index].set_radius(radius)

    def get_type_name(self, index):
        """Function to get name of atom type.

        Parameters
        ----------
        index : int
            Index of atom type.

        Returns
        -------
        output : str or None
            Name of atom type if it exists, else index.

        """
        if index >= self.n_types():
            return str(index)
        name = self.type_name[index]
        return str(index) if name is None else name

    def convert_fractional_to_cartesian(self, x):
        """Function to convert fractional coordinates to Cartesian.

        Parameters
        ----------
        x : array-like
            Fractional coordinates.

        Returns
        -------
        output : array-like
            Cartesian coordinates.

        """

        return np.array([np.dot(x, y) for y in self.simulation_cell])

    def convert_cartesian_to_fractional(self, x):
        """Function to convert Cartesian coordinates to fractional.

        Parameters
        ----------
        x : array-like
            Cartesian coordinates.

        Returns
        -------
        output : array-like
            Fractional coordinates.

        """

        return np.array([y.dot(x) for y in self.inverse_cell])

    def get_periodic_image(self, position, x, y, z):
        """Function to compute the position periodic image given its position in
        Cartesian coordinates.

        Parameters
        ----------
        position : array-like
            A list containing the Cartesian coordinates.
        x : float
            Number of steps in the X direction.
        y : float
            Number of steps in the Y direction.
        z : float
            Number of steps in the Z direction.

        Returns
        -------
        output : array-like
            New position.

        """

        l = np.array([x, y, z])
        output = position + np.array([l.dot(y) for y in self.simulation_cell])
        return output


    def get_minimum_distance(self, point1=None, point2=None, center=None,
                             neighbor=None):
        """Function to compute the minimum distance between any images of two
        points or to get the closest image of a neighboring atom to a given
        atom.

        Minimum distance algorithm:
        Compute the displacement between the points in Cartesian units.
        For each lattice vector:
            Compute the projection of the displacement onto this vector
            Divide that distance bvy the length of the lattice vector, and round
            to get the number of extra lattice vectors away this displacement
            vector is from the zero vector.
            Multiply the lattice vector by that factor and subtract the
            result from the current lattice vector to get a candidate
            displacement vector.
            If this vector is shorter than the current displacement vector,
            replace this vector with the new one.
        Compute the distance of this displacement vector.

        To get the closest image, we follow the exact same steps as computing
        the minimum distance. Except in the end, where we return the closest
        image.

        Parameters
        ----------
        point1 : array-like
            Fractional coordinates of point1. (Default value = None)
        point2 : array-like
            Fractional coordinates of point2. (Default value = None)
        center : int
            ID of central atom. (Default value = None)
        neighbor : int
            Neighboring atom. (Default value = None)

        Returns
        -------
        output : float or AtomImage
            Depending on the choice, either the minimum distance or the
            closest image.

        """

        disp = None
        subtract_pos = None
        flag = False
        image = None
        neighbor_atom = None

        if point1 is not None and point2 is not None:
            # Get the distance in cartesian units.
            disp = self.convert_fractional_to_cartesian(point2)
            subtract_pos = self.convert_cartesian_to_fractional(point1)
        elif center is not None and neighbor is not None:
            flag = True
            image = np.zeros(3, dtype=int)
            center_atom = self.get_atom(center)
            neighbor_atom = self.get_atom(neighbor)
            # Compute the displacement vector between these atoms.
            disp = neighbor_atom.get_position_cartesian().copy()
            subtract_pos = center_atom.get_position_cartesian()


        disp -= subtract_pos

        lat_vec = self.get_lattice_vectors()

        # For each direction, find the closest image.
        cur_dist = disp[0]**2 + disp[1]**2 + disp[2]**2
        for d in range(3):
            proj_D = disp[0] * lat_vec[d, 0] + disp[1] * lat_vec[d, 1] + disp[
                    2] * lat_vec[d, 2]
            proj_D /= lat_vec[d, 0]**2 + lat_vec[d, 1]**2 + lat_vec[d, 2]**2
            n_steps = int(round(proj_D))
            # Java's Math.round and python's np.math.round are slightly different.
            # For instance, the result of rounding -23.5 is -23 in java and
            # -24 in python. From python docs of the round function:
            # https://docs.python.org/2/library/functions.html#round
            # if two multiples are equally close, rounding is done away from 0
            # (so, for example, round(0.5) is 1.0 and round(-0.5) is -1.0).
            # In order to emulate the correct behavior, check if it the
            # argument is negative and ends with ".5". If so, add 1 to the
            # result of the int(round(x)) operation.
            if proj_D < 0 and str(proj_D).rstrip("0").endswith(".5"):
                n_steps += 1
            new_disp = disp - n_steps * lat_vec[d]
            new_dist = new_disp[0] ** 2 + new_disp[1] ** 2 + new_disp[2] ** 2
            if new_dist < cur_dist:
                if flag:
                    image[d] = -1 * n_steps
                disp = new_disp
                cur_dist = new_dist

        return np.math.sqrt(cur_dist) if not flag else AtomImage(neighbor_atom,
                                                              image)


import unittest

from cheml.search import GA_DEAP

Weights = (1,)
chromosome_length = 2
chromosome_type = (1,1)
bit_limits = ( (0,10), (0,5) )
mut_int_lower = (0,0)
mut_int_upper = (10,5)

def Evaluate(individual):
    return sum(individual),


class TestGA_DEAP(unittest.TestCase):
    def test_algorithm1(self):

        ga_search = GA_DEAP(Evaluate, Weights=Weights, chromosome_length = chromosome_length, chromosome_type = chromosome_type,
                       bit_limits = bit_limits, mut_int_lower=mut_int_lower, mut_int_upper=mut_int_upper )

        ga_search.fit()
        best_ind_df, best_individual = ga_search.algorithm_1()
        self.assertEqual(15, sum(best_individual))

    def test_algorithm2(self):
        ga_search = GA_DEAP(Evaluate, Weights=Weights, chromosome_length=chromosome_length, chromosome_type=chromosome_type,
                            bit_limits=bit_limits, mut_int_lower=mut_int_lower, mut_int_upper=mut_int_upper)
        ga_search.fit()
        best_ind_df, best_individual = ga_search.algorithm_2()
        self.assertEqual(15, sum(best_individual))

    def test_algorithm3(self):
        ga_search = GA_DEAP(Evaluate, Weights=Weights, chromosome_length=chromosome_length, chromosome_type=chromosome_type,
                            bit_limits=bit_limits, mut_int_lower=mut_int_lower, mut_int_upper=mut_int_upper)

        ga_search.fit()
        best_ind_df, best_individual = ga_search.algorithm_3()
        self.assertEqual(15, sum(best_individual))

    def test_algorithm4(self):
        ga_search = GA_DEAP(Evaluate, Weights=Weights, chromosome_length=chromosome_length, chromosome_type=chromosome_type,
                            bit_limits=bit_limits, mut_int_lower=mut_int_lower, mut_int_upper=mut_int_upper)

        ga_search.fit()
        best_ind_df, best_individual = ga_search.algorithm_4()
        self.assertEqual(15, sum(best_individual))





import unittest
import os
import shutil, tempfile
import pkg_resources

from cheml.wrappers.engine import run as wrapperRUN

CONFIG_PATH = pkg_resources.resource_filename('cheml', os.path.join('tests', 'configfiles'))


class Testwrapper(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)

    def test_template1(self):
        from cheml.notebooks.templates import template1
        script =  template1()
        wrapperRUN(script,os.path.join(self.test_dir,'test.out'))

    def test_template2(self):
        from cheml.notebooks.templates import template2
        script = template2()
        wrapperRUN(script, os.path.join(self.test_dir,'test.out'))

    def test_template3(self):
        from cheml.notebooks.templates import template3
        script = template3()
        # require rdkit
        # wrapperRUN(script, os.path.join(self.test_dir,'test.out'))
    
    def test_template4(self):
        from cheml.notebooks.templates import template4
        script = template4()
        # require Dragon
        #wrapperRUN(script, os.path.join(self.test_dir,'test.out'))
    
    def test_template5(self):
        from cheml.notebooks.templates import template5
        script = template5()
        wrapperRUN(script, os.path.join(self.test_dir,'test.out'))
    
    def test_template6(self):
        from cheml.notebooks.templates import template6
        script = template6()
        # runs slow
        #wrapperRUN(script, os.path.join(self.test_dir,'test.out'))
    
    def test_template7(self):
        from cheml.notebooks.templates import template7
        script = template7()
        wrapperRUN(script, os.path.join(self.test_dir,'test.out'))
    
    def test_template8(self):
        from cheml.notebooks.templates import template8
        script = template8()
        wrapperRUN(script, os.path.join(self.test_dir,'test.out'))
    
    def test_template9(self):
        from cheml.notebooks.templates import template9
        script = template9()
        wrapperRUN(script, os.path.join(self.test_dir,'test.out'))
    

    def test_template11(self):
        from cheml.notebooks.templates import template11
        script = template11()
        wrapperRUN(script, os.path.join(self.test_dir,'test.out'))

    def test_template12(self):
        from cheml.notebooks.templates import template12
        script = template12()
        wrapperRUN(script, os.path.join(self.test_dir,'test.out'))

    def test_test1(self):
        wrapperRUN(os.path.join(CONFIG_PATH,'test1.txt'), os.path.join(self.test_dir,'test.out'))


if __name__== '__main__':
    unittest.main()






import math
import random
import numpy as np
import copy
import pickle as pkl
from ..nn.nn_psgd import
class GA_nn(object):
    """
    Genetic Algorithm mixed with a trained neural network as its objective function

    Parameters
    ----------
    nnet: string
        The path to the pickle file that contains the weights and activation functions of a
        trained neural network.
    target: list
        list of desired values for outputs
    nVar: integer
        Number of elements in each chromosome
    normalizer: list, optional (default = [0,0])
        A list of standard deviation (first element) and mean (second element) of target values
        in the training set.
    crossover: string, optinal (default = 'Uniform')
        A choice of three different crossover methods: 'SinglePoint', 'DoublePoint', 'Uniform'
    selection: string, optional (default = 'RWS')
        A choice of three different methods for the selection of parent chromosomes:
            'RWS' : Roulette Wheel Selection
            'TS' : Tournament Selection
            'RS' : Random Selection
    selection_pressure: integer, optional (default = 3)
        selection pressure determines the effect of cost of a chromosome in the chance of selection
    init_pop: integer, optional (default = 100)
        The initialized population of chromosomes
    crossover_ratio: float in the range(0,1), optional (default = 0.8)
        The proportion of total number of chromosomes to be selected for crossover
    mutation_ratio: float in the range(0,1), optional (default = 0.3)
        The proportion of total number of chromosomes to be selected for mutation
    mutaton_rate: float, optional (default = 0.02)
        The proportion of total number of elements in a chromosome for mutation.
    tournament_size: integer, optional (default = 3)
        size of subgroups for the tournament selection
    n_iterations: integer, optional (default = 1000)
        Total number of iteration
    """
    def __init__(self, nnet, target, nVar, normalizer = [0,0], crossover = 'Uniform', selection = 'RWS',
                 selection_pressure = 3, init_pop=100, crossover_ratio = 0.8, mutation_ratio = 0.3,
                 mutation_rate = 0.02, tournament_size = 3,n_iterations = 1000):
        self.nnet = nnet
        self.target = target
        self.normalizer = normalizer
        self.nVar = nVar3366
        self.crossover = crossover
        self.selection = selection
        self.selection_pressure = selection_pressure
        self.init_pop = init_pop
        self.crossover_ratio = crossover_ratio
        self.mutation_ratio = mutation_ratio
        self.mutation_rate = mutation_rate
        self.tournament_size = tournament_size
        self.n_iterations = n_iterations

    def SinglePointCrossover(self, x1, x2):
        nVar=len(x1)
        c = random.randint(1,nVar-1)
        y1=x1[0:c]				# the start point and length of range(end-start) is important
        y1=y1+x2[c:nVar]      # the start point and length of range(end-start) is important
        y2=x2[0:c]
        y2=y2+x1[c:nVar]
        return tuple(y1), tuple(y2)

    def DoublePointCrossover(self, x1, x2):
        nVar = len(x1)
        cc = random.sample(range(1,nVar), 2)   # the start point and length of range(end-start) is important => here range is 1 to nVar-2
        c1 = min(cc)
        c2 = max(cc)
        y1 = x1[0:c1]+x2[c1:c2]+x1[c2:nVar]				# the start point and length of range(end-start) is important
        y2 = x2[0:c1]+x1[c1:c2]+x2[c2:nVar]      # the start point and length of range(end-start) is important
        return tuple(y1), tuple(y2)

    def UniformCrossover(self, x1, x2):
        nVar = len(x1)
        a = []
        for i in range (0,nVar):
            q = random.randint(0,1)
            a.append(q)
        alpha = np.array(a)
        y1 = alpha*x1+(1-alpha)*x2
        y2 = alpha*x2+(1-alpha)*x1
        return tuple(y1), tuple(y2)

    def RouletteWheelSelection(self, X):
        r = random.uniform(0,1)
        c=np.cumsum(X)
        c=np.array(c).tolist()
        f=0
        for i in range (0,len(c)-1):
            if r<=c[i]:
                f=i
                break
        return(f)

    def TournamentSelection(self, C, m):
        nVar=len(C)
        c=[]
        S = random.sample(range(0,nVar), m)
        for i in range (0,m):
            c.append(C[S[i]])
        c, S = (list(t)for t in zip(*sorted(zip(c, S))))
        return(S[0])

    def Mutate(self, x):
        nVar = len(x)
        nmu = math.ceil(self.mutation_rate*nVar)
        j = random.sample(range(nVar), int(nmu))
        y = copy.deepcopy(x)
        y = list(y)
        for i in xrange(int(nmu)):
            y[j[i]] = 1-y[j[i]]
        return tuple(y)

    def CostFunction(self,x):
        with open(self.nnet, 'rb') as network:
            self.nnet = pkl.load(network)
        prediction = nn.output(np.array(x),nnet)
        prediction = (self.normalizer[0] * prediction) + self.normalizer[1] #(0:std, 1:mean)
        error = np.mean(np.abs(prediction - np.array(self.target)))
        return error

    def Crossover(self,P):
        nc = 2 * math.ceil(self.crossover_ratio * self.init_pop / 2)
        popc = []
        costc = []
        for k in xrange(0, int(nc / 2)):
            if self.selection == 'RWS':
                q = self.RouletteWheelSelection(P)  # SP=0 is like random
                r = self.RouletteWheelSelection(P)  # SP=0 is like random
            elif self.selection == 'TS':
                q = self.TournamentSelection(self.cost, self.tournament_size)
                r = self.TournamentSelection(self.cost, self.tournament_size)
            elif self.selection == 'RS':
                q = random.randint(0,nPop-1)
                r = random.randint(0,nPop-1)
            else:
                msg = "'%s' is not a valid selection method"%self.selection
                raise NameError(msg)
            # Parents
            p1 = pop[q]
            p2 = pop[r]

            # Apply Crossover & Make popc
            if self.crossover == 'SinglePoint':
                a = self.SinglePointCrossover(p1,p2)       # SinglePointCrossover SPC
            elif self.crossover == 'DoublePoint':
                a = self.DoublePointCrossover(p1,p2)       # DoublePointCrossover DPC
            elif self.crossover == 'Uniform':
                a = self.UniformCrossover(np.array(p1), np.array(p2))          # UniformCrossover UC

            # Evaluate Offsprings & Make costc
            a1 = self.repair(a[0])
            popc.append(a1)
            costc.append(self.CostFunction(a1))
            a2 = self.repair(a[1])
            popc.append(a2)
            costc.append(self.CostFunction(a2))
        return popc, costc

    def Mutation(self,P):
        nm = math.ceil(self.mutation_ratio * self.init_pop)
        popm = []
        costm = []
        for k in range(0, int(nm)):
            if self.selection == 'RWS':
                q = self.RouletteWheelSelection(P)
            elif self.selection == 'TS':
                q = self.TournamentSelection(self.cost, self.tournament_size)
            elif self.selection == 'RS':
                q = random.randint(0,nPop-1)
            else:
                msg = "'%s' is not a valid selection method"%self.selection
                raise NameError(msg)
            # Mutant
            p = self.pop[q]

            # Apply Mutation & Make popm
            a = self.Mutate(p)
            a = self.repair(a)
            popm.append(a)

            # Evaluate Mutant and Make costm
            costm.append(self.CostFunction(np.array(a)))
        return popm, costm

    def fit(self):
        self.pop=[]
        self.cost=[]
        for i in xrange (0,self.init_pop):
            # Initialize Position
            a=[]
            for i in range (0,nvar):
                q = random.randint(0,1)
                a.append(q)
            a = tuple(a)
            self.pop.append(a)

            # Evaluation
            self.cost.append(self.CostFunction(a))

        ## Sort Population
        self.cost, self.pop = (list(t)for t in zip(*sorted(zip(self.cost, self.pop))))

        ## Store cost
        worstcost=[]
        worstcost.append(cost[-1])
        WorstCost=np.nanmax(worstcost)

        ## Main Loop
        for it in xrange (0,self.n_iterations):
            # Calculate Selection Probabilities
            P = np.exp(-self.selection_pressure*np.array(cost)/WorstCost)
            P = P / float(sum(P))
            popc, costc = self.Crossover(P)
            popm, costm = self.Mutation(P)
            # Create Merge Population & Cost
            pop = pop + popc + popm
            cost = cost + costc + costm

            pop = tuple(pop)
            cost = tuple(cost)

            # Sort Population
            cost, pop = (list(t)for t in zip(*sorted(zip(cost, pop))))

            # Update WorstCost
            worstcost.append(cost[-1])
            WorstCost = np.nanmax(worstcost)

            # Truncation
            pop = pop[0:nPop]
            cost = cost[0:nPop]

            # Store Best Solution Ever Found
            BestSol = pop[0]

            # Store Best Cost Ever Found
            BestCost = cost[0]


            # Save Iteration Information
            with open('GA_Iteration_Records.txt','a') as it_file:
                it_file.write('Iteration: %i, Best cost $ %s\n'%(it, str(BestCost)))
                it_file.write('Iteration: %i, Best chromosome: %s\n'%(it, str(BestSol)))
                it_file.write('          GA+NN by MHL\n')







import unittest
from .....vassal.analysis.voronoi.VoronoiEdge import VoronoiEdge
from .....vassal.analysis.voronoi.VoronoiFace import VoronoiFace
from .....vassal.analysis.voronoi.VoronoiVertex import VoronoiVertex
from .....vassal.data.Atom import Atom
from .....vassal.data.AtomImage import AtomImage
from .....vassal.data.Cell import Cell

class testVoronoiVertex(unittest.TestCase):
    def test_creation(self):
        # Make a simple crystal.
        cell = Cell()
        cell.add_atom(Atom([0, 0, 0], 0))

        # Initialize faces.
        image = AtomImage(cell.get_atom(0), [0, 0, 1])
        face1 = VoronoiFace(cell.get_atom(0), image, radical=True)
        image = AtomImage(cell.get_atom(0), [0, 1, 0])
        face2 = VoronoiFace(cell.get_atom(0), image, radical=True)
        image = AtomImage(cell.get_atom(0), [1, 0, 0])
        face3 = VoronoiFace(cell.get_atom(0), image, radical=True)

        # Create edges.
        edge1 = VoronoiEdge(face1, face2)
        edge2 = VoronoiEdge(face1, face3)

        # Create vertex.
        vertex = VoronoiVertex(edge1=edge1, edge2=edge2)

        # Test properties.
        self.assertEquals(edge2, vertex.get_previous_edge())
        self.assertEquals(edge1, vertex.get_next_edge())

        # Make sure the order of edges on creation doesn't matter.
        self.assertEquals(vertex, VoronoiVertex(edge1=edge2, edge2=edge1))

        # Create a new vertex, ensure that it is different.
        edge3 = VoronoiEdge(face2, face3)
        self.assertFalse(vertex.__eq__(VoronoiVertex(edge1=edge3,
                                                     edge2=edge2)))


# coding=utf-8
import numpy as np
import pandas as pd
import types
from ....data.materials.CrystalStructureEntry import CrystalStructureEntry
from ....data.materials.util.LookUpData import LookUpData
from ....models.regression.crystal.PRDFRegression import PRDFRegression

class PRDFAttributeGenerator:
    """Class to compute attributes based on the Pair Radial Distribution
    Function (PRDF).

    Based on work by Schutt et al. [1].

    Attributes
    ----------
    cut_off_distance : float
        Cutoff distance for PRDF.
    n_points : int
        Number of distance points to evaluate.
    element_list : array-like
        Elements to use in PRDF. A list of int values.

    References
    ----------
    .. [1] K. T. Schütt, H. Glawe, F. Brockherde, A. Sanna, K. R. Müller,
    and E. K. U. Gross, "How to represent crystal structures for machine
    learning: Towards fast prediction of electronic properties," Physical
    Review B, vol. 89, no. 20, May 2014.

    """

    def __init__(self):
        """Function to create instance and initialize fields.

        """

        # Cutoff distance for PRDF.
        self.cut_off_distance = 10

        # Number of distance points to evaluate.
        self.n_points = 20

        # List of Elements to use in PRDF.
        self.element_list = []

    def set_cut_off_distance(self, d):
        """Function to set the maximum distance to consider when computing the
        PRDF.

        Parameters
        ----------
        d : float
            Desired cutoff distance.

        """

        self.cut_off_distance = d

    def set_n_points(self, n_p):
        """Function to set the number of points on each PRDF to store.

        Parameters
        ----------
        n_p : int
            Number of evaluation points.

        """

        self.n_points = n_p

    def clear_element_list(self):
        """Function to clear out the elements in element list.

        """

        self.element_list = []

    def set_elements(self, entries):
        """Function to set the elements when computing PRDF.

        Parameters
        ----------
        data : array-like
            A list of CompositionEntry's containing each element to be added.

        """

        self.clear_element_list()
        for entry in entries:
            for elem in entry.get_element_ids():
                if elem not in self.element_list:
                    self.add_element(id=elem)

    def add_element(self, id=None, name=None):
        """Function to add element to list used when computing PRDF.

        Parameters
        ----------
        id   : int
            ID of element (Atomic number - 1).
        name : str
            Name of the element.

        Raises
        ------
        ValueError
            If both arguments are None.
            If entered element name can not be found in database.

        """

        if id is not None:
            self.element_list.append(id)
        elif name is None:
            raise ValueError("Either id or name must be provided to locate "
                             "element to be added.")
        elif name not in LookUpData.element_names:
            raise ValueError("No such element: "+name)
        else:
            self.element_list.append(LookUpData.element_names.index(name))

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Crystal structures for which features are to be generated. A list
            of CrystalStructureEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CrystalStructureEntry instances.

        """

        # Initialize list of feature values for pandas data frame.
        feat_values = []

        # Raise exception if input argument is not of type list of
        # CrystalStructureEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")
        elif (entries and not isinstance(entries[0], CrystalStructureEntry)):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")

        # Get names of elements in set.
        self.element_list.sort()
        elem_names = [LookUpData.element_names[i] for i in self.element_list]
        elem_ids = list(self.element_list)

        # Get the step size for the PRDF.
        step_size = self.cut_off_distance / self.n_points

        # Insert header names here.
        feat_headers = ["{:s}_{:s}_R={:3f}".format(elem_a, elem_b, step *
                        step_size) for elem_a in elem_names for elem_b in
                        elem_names for step in range(1, self.n_points + 1)]

        # Initialize PRDF tool.
        tool = PRDFRegression()
        tool.set_cut_off(self.cut_off_distance)
        tool.set_n_bins(self.n_points)

        l_fh = len(feat_headers)
        # Generate features for each entry.
        for entry in entries:
            # Compute the PRDF.
            prdf = tool.compute_representation(entry.get_structure())
            tmp_array = np.zeros(l_fh, dtype=float)
            for pair in prdf:
                # Determine position in output.
                elem_a = elem_ids.index(pair[0])
                elem_b = elem_ids.index(pair[1])

                prdf_pos = (elem_a * len(self.element_list) + elem_b) * \
                           self.n_points
                tmp_array[prdf_pos: (prdf_pos + self.n_points)] = prdf[pair][
                                    0 : self.n_points]

            feat_values.append(tmp_array)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import unittest
from numpy.linalg import norm
from ....vassal.analysis.PairDistanceAnalysis import PairDistanceAnalysis
from ....vassal.data.Atom import Atom
from ....vassal.data.Cell import Cell

class testPairDistanceAnalysis(unittest.TestCase):
    def setUp(self):
        self.structure = Cell()
        self.structure.set_basis(lengths=[1, 1, 1], angles=[90, 90, 90])
        self.structure.add_atom(Atom([0, 0, 0], 0))
        self.pda = PairDistanceAnalysis()
        self.pda.analyze_structure(self.structure)

    def tearDown(self):
        self.structure = None
        self.pda = None

    def test_get_all_neighbors_of_atom(self):
        # With orthorhombic basis.
        self.pda.set_cutoff_distance(1.1)
        output = self.pda.get_all_neighbors_of_atom(0)
        self.assertEquals(6, len(output))
        self.assertAlmostEquals(1.0, output[0][1], delta=1e-6)

        # Adding a second atom.
        self.structure.add_atom(Atom([0.5, 0.5, 0.5], 0))
        output = self.pda.get_all_neighbors_of_atom(0)
        self.assertEquals(14, len(output))

        # Altering the basis to something weird.
        new_basis = self.structure.get_basis()
        new_basis[1][0] = 14
        output = self.pda.get_all_neighbors_of_atom(0)
        self.assertEquals(14, len(output))

        # Check that images match up.
        center_pos = self.structure.get_atom(0).get_position_cartesian()
        for image in output:
            v = image[0].get_position() - center_pos
            self.assertAlmostEquals(image[1], norm(v), delta=1e-6)

    def test_PRDF(self):
        # With orthorhombic basis.
        self.pda.set_cutoff_distance(2.1)

        # Run code.
        prdf = self.pda.compute_PRDF(50)
        self.assertEquals(1, len(prdf))
        self.assertEquals(1, len(prdf[0]))
        self.assertEquals(50, len(prdf[0][0]))

        # Make sure that it finds 4 peaks.
        n_peaks = 0
        for val in prdf[0][0]:
            if val > 0:
                n_peaks += 1

        self.assertEquals(4, n_peaks)

        # Add another atom, repeat.
        self.structure.add_atom(Atom([0.5, 0.5, 0.5], 1))

        # Run again.
        prdf = self.pda.compute_PRDF(50)
        self.assertEquals(2, len(prdf))
        self.assertEquals(2, len(prdf[0]))
        self.assertEquals(50, len(prdf[0][0]))

        # Make sure A-B prdf has 2 peaks.
        n_peaks = 0
        for val in prdf[0][1]:
            if val > 0:
                n_peaks += 1

        self.assertEquals(2, n_peaks)

        # Increase basis.
        self.structure.set_basis(lengths=[2, 2, 2], angles=[90, 90, 90])
        self.pda.analyze_structure(self.structure)

        # Run again.
        prdf = self.pda.compute_PRDF(50)
        self.assertEquals(2, len(prdf))
        self.assertEquals(2, len(prdf[0]))
        self.assertEquals(50, len(prdf[0][0]))

        # Make sure A-B prdf has 1 peaks.
        n_peaks = 0
        for val in prdf[0][1]:
            if val > 0:
                n_peaks += 1

        self.assertEquals(1, n_peaks)


import unittest
import numpy.testing as np
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.LookUpData import LookUpData
from ....utility.tools.OxidationStateGuesser import OxidationStateGuesser

class testOxidationStateGuesser(unittest.TestCase):
    def test_guesser(self):
        ox = OxidationStateGuesser()
        ox.set_oxidationstates(LookUpData.load_property("OxidationStates"))
        ox.set_electronegativity(LookUpData.load_property("Electronegativity"))

        res = ox.get_possible_states(CompositionEntry(composition="NaCl"))
        self.assertEquals(1, len(res))
        np.assert_array_equal([1, -1], res[0])

        res = ox.get_possible_states(CompositionEntry(composition="Fe2O3"))
        self.assertEquals(1, len(res))
        np.assert_array_equal([3, -2], res[0])

        res = ox.get_possible_states(CompositionEntry(composition="NaHCO3"))
        self.assertEquals(1, len(res))
        np.assert_array_equal([10, 0, 5, 7], CompositionEntry(
            composition="NaHCO3").get_element_ids())
        np.assert_array_equal([1, 1, 4, -2], res[0])

        res = ox.get_possible_states(CompositionEntry(composition="NH3"))
        self.assertEquals(2, len(res))
        np.assert_array_equal([0, 6], CompositionEntry(
            composition="NH3").get_element_ids())
        np.assert_array_equal([1, -3], res[0])

        res = ox.get_possible_states(CompositionEntry(composition="NaAl"))
        self.assertEquals(0, len(res))

        res = ox.get_possible_states(CompositionEntry(composition="PbTiO3"))
        self.assertEquals(2, len(res))
        np.assert_array_equal([21, 81, 7], CompositionEntry(
            composition="PbTiO3").get_element_ids())
        np.assert_array_equal([4, 2, -2], res[0])


import numpy as np
from numpy.linalg import norm
from ..voronoi.VoronoiVertex import VoronoiVertex

class VoronoiEdge:
    """Class describing an edge of a cell in a Voronoi tessellation.

    Attributes
    ----------
    edge_face : Plane
        Plane defining one face containing this edge.
    intersecting_face : Plane
        Plane defining the other face defining this edge.
    line : Line
        Line representing this edge.
    direction : array-like
        Direction of edge.
    beginning : float
        Point marking the beginning of this edge.
    end : float
        Point marking the end of this edge.
    next_edge : VoronoiEdge
        Edge after this edge.
    previous_edge : VoronoiEdge
        Edge before this edge.
    """

    def __init__(self, edge_face=None, intersecting_face=None):
        """Function to initialize an edge instance.

        Parameters
        ----------
        edge_face : Plane
            Plane defining one face containing this edge.
        intersecting_face : Plane
            Plane defining the other face defining this edge.

        Raises
        ------
        Exception
            Planes are parallel.
        """

        # EdgeFace on which this edge exists.
        self.edge_face = edge_face

        # EdgeFace that intersected edge_face to form this edge.
        self.intersecting_face = intersecting_face

        # Line representing this edge.
        self.line = None

        # Direction of edge.
        self.direction = None

        # Point marking the beginning of this edge.
        self.beginning = -float("inf")

        # Edge after this edge.
        self.next_edge = None

        # Point marking the end of this edge.
        self.end = float("inf")

        # Edge before this edge.
        self.previous_edge = None

        # Compute the line.
        self.line = edge_face.get_plane().intersection(other=
            intersecting_face.get_plane())
        if self.line is None:
            raise Exception("Planes are parallel.")

        # Ensure vector is CCW w.r.t edge face.
        cut_direction = -intersecting_face.get_normal()
        self.direction = self.line.get_direction()
        if not self.is_ccw(vec1=edge_face.get_normal(),
                           vec2=self.direction, vec3=cut_direction):
            self.line = self.line.revert()

    def is_ccw(self, vec1=None, vec2=None, vec3=None, edge2=None):
        """Function to compute whether, given a face normal, edges with two
        directions are CCW.

        Parameters
        ----------
        vec1 : array-like
            Face normal. (Default value = None)
        vec2 : array-like
            Direction of edge a. (Default value = None)
        vec3 : array-like
            Direction of edge b. (Default value = None)
        edge2 : VoronoiEdge
            Edge 2. (Default value = None)

        Returns
        -------
        type : bool
            Whether they are CCW.

        """
        v0 = vec1
        v1 = vec2
        v2 = vec3
        if edge2 is not None:
            v0 = self.edge_face.get_normal()
            v1 = self.direction
            v2 = edge2.direction

        term1 = v0[0] * v1[1] * v2[2] - v0[0] * v1[2] * v2[1]
        term2 = v0[1] * v1[2] * v2[0] - v0[1] * v1[0] * v2[2]
        term3 = v0[2] * v1[0] * v2[1] - v0[2] * v1[1] * v2[0]

        ccw = term1 + term2 + term3

        # Check if the magnitude of ccw is less than machine epsilon. If so,
        # they are considered to be collinear. Return False.
        if abs(ccw) <= np.finfo(float).eps:
            return False

        return ccw > 0

    @classmethod
    def compute_intersection(self, edge1, edge2, just_join=False):
        """Function to compute intersection between two edges or join them.

        Computes the intersection between two vectors. If they intersect,
        within the bounds of each edge, set new boundaries for each edge and
        change the next and previous edge as appropriate.

        Will check orientation of edges to see if edge1 is the edge following
        edge2 (or vis versa), and join them accordingly.

        Parameters
        ----------
        edge1 : VoronoiEdge
            First edge.
        edge2 : VoronoiEdge
            Second edge.
        just_join : bool
            Whether to just join them or compute intersection.
            (Default value = False)

        Returns
        -------
        output : bool or None
            Whether two edges intersect or nothing depending on the input.

        Raises
        ------
        Exception
            Edges do not intersect.
        """

        # Determine the point at which the edges intersect.
        point = edge1.line.intersection(edge2.line)
        if point is None:
            if just_join:
                raise Exception("Edges do not intersect.")
            else:
                return False

        # Determine the relationship between edges (using their directions).
        is_forward = edge1.is_ccw(edge2=edge2)

        # Using the direction, check whether intersection is within bounds of
        #  each edge.
        edge1_terminus = edge1.line.get_abscissa(point)
        edge2_terminus = edge2.line.get_abscissa(point)

        if not just_join:
            within_bounds = None
            if is_forward:
                within_bounds = edge1_terminus < edge1.end and edge2_terminus > \
                                                               edge2.beginning
            else:
                within_bounds = edge1_terminus > edge1.beginning and \
                                edge2_terminus < edge2.end

            if not within_bounds:
                return False

        # Now update the edges accordingly.
        if is_forward:
            edge1.end = edge1_terminus
            edge1.next_edge = edge2
            edge2.beginning = edge2_terminus
            edge2.previous_edge = edge1
        else:
            edge1.beginning = edge1_terminus
            edge1.previous_edge = edge2
            edge2.end = edge2_terminus
            edge2.next_edge = edge1

        if not just_join:
            return True

    def __eq__(self, other):
        """Function to check if instance is equal to another edge.

        Parameters
        ----------
        other : VoronoiEdge
            Other edge.

        Returns
        -------
        type : bool
            True if instance equal to other, else False.
        """
        if isinstance(other, VoronoiEdge):
            return other.edge_face.__eq__(self.edge_face) and \
                   other.intersecting_face.__eq__(self.intersecting_face)
        return False

    def __hash__(self):
        """Function to compute the hash value of instance.

        Returns
        -------
        type : int
            Hash value.
        """
        h = 5
        h = 43 * h + id(self.edge_face)
        h = 43 * h + id(self.intersecting_face)
        return h

    def __cmp__(self, other):
        """Function to compare instance and another edge.

        Parameters
        ----------
        other : VoronoiEdge
            Other edge.

        Returns
        -------
        type : int
            -1 if instance < other, +1 if instance > other, else 0.
        """
        if self.edge_face.__eq__(other.edge_face):
            return self.intersecting_face.__cmp__(other.intersecting_face)
        return self.edge_face.__cmp__(other.edge_face)

    def get_line(self):
        """Function to get the line defining this edge.

        Returns
        -------
        line : Line
            Line representing this edge.

        """
        return self.line

    def get_edge_face(self):
        """Function to get the face containing this edge.

        Returns
        -------
        edge_face : Plane
            Face containing this edge.
        """
        return self.edge_face

    def get_intersecting_face(self):
        """Function to get the other face associated with this edge.

        Returns
        -------
        intersecting_face : Plane
            Other face associated with this edge.
        """
        return self.intersecting_face

    def get_next_edge(self):
        """Function to get the next edge on this face.

        Returns
        -------
        next_edge : VoronoiEdge
            Next edge (None if no such edge).

        """
        return self.next_edge

    def get_previous_edge(self):
        """Function to get the previous edge on this face.

        Returns
        -------
        previous_edge : VoronoiEdge
            Previous edge (None if no such edge).

        """
        return self.previous_edge

    def get_length(self):
        """Function to get the length of this edge.

        Returns
        -------
        type : float
            Length.
        """
        return norm((self.beginning - self.end) * self.direction)

    def __str__(self):
        """Function to get the string representation of instance.

        Returns
        -------
        output : str
            String representation of instance.

        """
        output = "("+self.edge_face.__str__()+"," \
                        ""+self.intersecting_face.__str__()+")"
        return output

    def find_next_edge(self, candidates):
        """Function to find the edge that is likely to be "next" on a face that
        contains this edge.

        This is computed by first computing all edges
        that are oriented CCW to this edge. Next, the edge that is closest to
        the origin is found. If multiple edges intersect at the same point,
        the one with the greatest angle between the direction of this edge is
        selected.

        Parameters
        ----------
        candidates : array-like
            All candidate edges.

        Returns
        -------
        choice : VoronoiEdge
            The next edge. None if no suitable candidate is found.

        """

        # Locate the ccw edges.
        ccw_edges = []
        for edge in candidates:
            # print self.intersecting_face.outside_atom.__str__(), \
            #     edge.intersecting_face.outside_atom.__str__()
            if not self.__eq__(edge):
                flag = self.is_ccw(edge2=edge)
                # print edge.intersecting_face.outside_atom, flag
                if flag:
                    ccw_edges.append(edge)

        # Check if any were found.
        if len(ccw_edges) == 0:
            return None
        elif len(ccw_edges) == 1:
            return ccw_edges[0]

        # Find the closest edge(s).
        closest_edges = []
        min_dist = float("inf")
        min_point = np.array([np.inf]*3)
        for edge in ccw_edges:
            other_line = edge.get_line()
            # If this line contains the minimum point, add it to list.
            if other_line.contains(min_point):
                closest_edges.append(edge)
                continue

            intersection = self.line.intersection(other_line)
            if intersection is None:
                # Line is anti parallel.
                continue

            # See if it is the closest.
            x = self.line.get_abscissa(intersection)
            if x < min_dist:
                closest_edges = []
                min_dist = x
                min_point = intersection
                closest_edges.append(edge)

        # If only one edge, return answer.
        if len(closest_edges) == 1:
            return closest_edges[0]

        # Otherwise, find the edge with the largest angle.
        max_angle = 0
        choice = None
        for edge in closest_edges:
            angle = self.angle_between(self.line.get_direction(),
                                       edge.get_line().get_direction())
            if angle > max_angle:
                choice = edge
                max_angle = angle

        return choice

    def get_start_vertex(self):
        """Function to get the vertex at the beginning of this vector.

        Returns
        -------
        type : VoronoiVertex
            Starting vertex.

        """
        return VoronoiVertex(edge1=self, edge2=self.previous_edge)

    def get_end_vertex(self):
        """Function to get the vertex at the end of this vector.

        Returns
        -------
        type : VoronoiVertex
            End vertex.

        """
        return VoronoiVertex(edge1=self, edge2=self.next_edge)

    def generate_pair(self):
        """Function to generate the edge that corresponds to this edge on the
        intersecting face.

        Returns
        -------
        type : VoronoiEdge
            Newly instantiated edge.

        Raises
        ------
        Exception
            If it is not possible to generate edge.

        """
        try:
            return VoronoiEdge(self.get_intersecting_face(),
                               self.get_edge_face())
        except Exception:
            raise Exception("Shouldn't be possible.")

    def angle_between(self, v1, v2):
        """Function to compute the angular separation between two vectors.

        Documentation obtained from:
        http://commons.apache.org/proper/commons-math/javadocs/api-3.3/org
        /apache/commons/math3/geometry/euclidean/threed/Vector3D.html

        This method computes the angular separation between two vectors using
        the dot product for well separated vectors and the cross product for
        almost aligned vectors. This allows to have a good accuracy in all
        cases, even for vectors very close to each other.

        Parameters
        ----------
        v1 : array-like
            First vector.
        v2 : array-like
            Second vector.

        Returns
        -------
        type : float
            Angular separation between v1 and v2.

        """
        norm_product = norm(v1) * norm(v2)

        if norm_product == 0:
            raise Exception("Norms are zero!")
        dp = np.dot(v1, v2)
        threshold = norm_product * 0.9999
        if dp < -threshold or dp > threshold:
            # The vectors are almost aligned, compute using the sine.
            v3 = np.cross(v1, v2)
            x = np.math.asin(norm(v3) / norm_product)
            if dp >= 0:
                return x
            return np.math.pi - x
        else:
            # The vectors are sufficiently separated to use the cosine.
            return np.math.acos(dp / norm_product)

    def print_properties(self):
        """Function to print different properties of the Voronoi Edge instance.

        Mainly used for debugging purposes.
        """
        print "Edge face:", self.edge_face.outside_atom.__str__(), \
            "Intersecting face:", self.intersecting_face.outside_atom.__str__()
        print "Line zero:", self.line.zero
        print "Line direction:", self.line.direction
        print "Line tolerance:", self.line.tolerance
        print "Edge direction:", self.direction
        print "Beginning:", self.beginning
        print "End:", self.end
        print